# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper functions for fetch source links."""

import re

from clusterfuzz._internal.base import utils

DRIVE_LETTER_REGEX = re.compile(r'^[a-zA-Z]:\\')
RANGE_LIMIT = 10000
SOURCE_START_ID = 'src/'
SOURCE_STRIP_REGEX = re.compile(r'^[/]?src[/]?')
STACK_FRAME_PATH_LINE_REGEX = re.compile(
    r'(?<=\[|\(|\s)([a-zA-Z/.][^\s]*?)\s*(:|@)\s*(\d+)(?=\]$|\)$|:\d+$|$)')


class ComponentPath(object):

  def __init__(self, source=None, relative_path=None, display_path=None):
    self.source = source
    self.relative_path = relative_path
    self.display_path = display_path

  def __eq__(self, other):
    return (self.source == other.source and
            self.relative_path == other.relative_path and
            self.display_path == other.display_path)


class VCSViewer(object):
  """Base viewer class."""
  VCS_URL_REGEX = None
  VCS_REVISION_SUB = None
  VCS_REVISION_DIFF_SUB = None
  VCS_REVISION_PATH_LINE_SUB = None

  def __init__(self, url):
    self.url = url

  def get_mapped_url(self, repl, **kwargs):
    """Return mapped url given a url map and arguments."""
    mapped_url = self.VCS_URL_REGEX.sub(repl, self.url)
    mapped_url = mapped_url.format(**kwargs)
    return mapped_url

  def get_source_url_for_revision(self, revision):
    """Return source revision url given a url and revision."""
    if not self.VCS_REVISION_SUB:
      return None

    return self.get_mapped_url(self.VCS_REVISION_SUB, revision=revision)

  def get_source_url_for_revision_diff(self, start_revision, end_revision):
    """Return source revision diff url given a url and revision."""
    if not self.VCS_REVISION_DIFF_SUB:
      return None

    return self.get_mapped_url(
        self.VCS_REVISION_DIFF_SUB,
        start_revision=start_revision,
        end_revision=end_revision,
        range_limit=RANGE_LIMIT)

  def get_source_url_for_revision_path_and_line(self, revision, path, line):
    """Return source revision url given a url, revision, path and line."""
    if not self.VCS_REVISION_PATH_LINE_SUB:
      return None

    return self.get_mapped_url(
        self.VCS_REVISION_PATH_LINE_SUB,
        revision=revision,
        path=path,
        line=line)


class FreeDesktopVCS(VCSViewer):
  VCS_URL_REGEX = re.compile(
      r'https://anongit\.freedesktop\.org/git/(.*)\.git$')
  VCS_REVISION_SUB = r'https://cgit.freedesktop.org/\1/commit/?id={revision}'
  VCS_REVISION_DIFF_SUB = (r'https://cgit.freedesktop.org/\1/diff/'
                           r'?id={end_revision}&id2={start_revision}')
  VCS_REVISION_PATH_LINE_SUB = (
      r'https://cgit.freedesktop.org/\1/tree/{path}?id={revision}#n{line}')


class GitHubVCS(VCSViewer):
  VCS_URL_REGEX = re.compile(r'(https://github\.com/(.*?))(\.git)?$')
  VCS_REVISION_SUB = r'\1/commit/{revision}'
  VCS_REVISION_DIFF_SUB = (r'\1/compare/{start_revision}...{end_revision}')
  VCS_REVISION_PATH_LINE_SUB = r'\1/blob/{revision}/{path}#L{line}'


class GitLabVCS(GitHubVCS):
  VCS_URL_REGEX = re.compile(
      r'(https://gitlab(\.[\w\.\-]+)?\.(com|org)/(.*?))(\.git)?$')


class GoogleSourceVCS(VCSViewer):
  VCS_URL_REGEX = re.compile(
      r'(https://[^/]+\.googlesource\.com/(.*?))(\.git)?$')
  VCS_REVISION_SUB = r'\1/+/{revision}'
  VCS_REVISION_DIFF_SUB = (
      r'\1/+log/{start_revision}..{end_revision}?pretty=fuller&n={range_limit}')
  VCS_REVISION_PATH_LINE_SUB = r'\1/+/{revision}/{path}#{line}'


class GoogleVCS(VCSViewer):
  VCS_URL_REGEX = re.compile(r'^//(.*)$')
  VCS_REVISION_SUB = r'https://cs.corp.google.com/\1/?rcl={revision}'
  VCS_REVISION_PATH_LINE_SUB = (
      r'https://cs.corp.google.com/\1/{path}?rcl={revision}&l={line}')


class MercurialVCS(VCSViewer):
  VCS_URL_REGEX = re.compile(r'(https?://hg\.(.*))')
  VCS_REVISION_SUB = r'\1/rev/{revision}'
  VCS_REVISION_DIFF_SUB = (r'\1/log?rev={start_revision}%3A%3A{end_revision}'
                           r'&revcount={range_limit}')
  VCS_REVISION_PATH_LINE_SUB = r'\1/file/{revision}/{path}#l{line}'


VCS_LIST = [
    FreeDesktopVCS,
    GitHubVCS,
    GitLabVCS,
    GoogleSourceVCS,
    GoogleVCS,
    MercurialVCS,
]


def get_component_source_and_relative_path(path, revisions_dict):
  """Get component source and relative path given a revisions dictionary and
  path."""
  if not revisions_dict:
    return ComponentPath()

  normalized_path = normalize_source_path(path)
  if normalized_path is None:
    return ComponentPath()

  component_sources = sorted(list(revisions_dict.keys()), key=len, reverse=True)
  default_component_source = None
  for component_source in component_sources:
    # Trailing slash is important so that we match the exact component source.
    # E.g. without slash, we would match src/webrtc_overrides with src/webrtc
    # which is incorrect.
    stripped_component_source = (
        SOURCE_STRIP_REGEX.sub('', component_source) + '/')

    if normalized_path.startswith(stripped_component_source):
      relative_path = utils.strip_from_left(normalized_path,
                                            stripped_component_source)
      return ComponentPath(component_source, relative_path, normalized_path)

    if stripped_component_source == '/':
      default_component_source = component_source

  if default_component_source is None:
    return ComponentPath()

  return ComponentPath(default_component_source, normalized_path,
                       normalized_path)


def get_vcs_viewer_for_url(url):
  """Return a VCS instance given an input url."""
  for vcs in VCS_LIST:
    if vcs.VCS_URL_REGEX.match(url):
      return vcs(url)

  return None


def linkify_stack_frame(stack_frame, revisions_dict):
  """Linkify a stack frame with source links to its repo."""
  match = STACK_FRAME_PATH_LINE_REGEX.search(stack_frame)
  if not match:
    # If this stack frame does not contain a path and line, bail out.
    return stack_frame

  path = match.group(1)
  line = match.group(3)

  component_path = get_component_source_and_relative_path(path, revisions_dict)
  if not component_path.source:
    # Can't find any matching component source in revisions dict, bail out.
    return stack_frame

  source_dict = revisions_dict[component_path.source]
  repo_url = source_dict['url']
  revision = source_dict['rev']
  vcs_viewer = get_vcs_viewer_for_url(repo_url)
  if not vcs_viewer:
    # If we don't support the vcs, bail out.
    return stack_frame

  link_html = r'<a href="{url}">{path}:{line}</a>'.format(
      url=vcs_viewer.get_source_url_for_revision_path_and_line(
          revision, component_path.relative_path, line),
      path=component_path.display_path,
      line=line)

  linkified_stack_frame = STACK_FRAME_PATH_LINE_REGEX.sub(
      link_html, stack_frame)

  return linkified_stack_frame


def normalize_source_path(path):
  """Normalizes source path for comparison with component sources."""
  # Account for ../../ at start of path due to working directory
  # out/<build_dir>/ at time of build generation (chromium only).
  path = utils.remove_prefix(path, '../../')

  # Remove /proc/self/cwd prefix added by Bazel.
  path = utils.remove_prefix(path, '/proc/self/cwd/')

  # Cross-platform way to determine path absoluteness.
  is_path_absolute = path.startswith('/') or DRIVE_LETTER_REGEX.match(path)

  # Normalize backslashes into slashes.
  normalized_path = path.replace('\\', '/')

  if is_path_absolute:
    source_start_id_index = normalized_path.find(SOURCE_START_ID)
    if source_start_id_index == -1:
      # This absolute path does not have source start id, so we cannot
      # figure out a relative path. Bail out.
      return None

    return normalized_path[source_start_id_index + len(SOURCE_START_ID):]

  return normalized_path



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Manage corpora."""

from flask import request
from google.cloud import ndb

from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from handlers import base_handler
from libs import form
from libs import handler
from libs import helpers


class Handler(base_handler.Handler):
  """Manage data bundles."""

  @handler.unsupported_on_local_server
  @handler.get(handler.HTML)
  @handler.check_admin_access_if_oss_fuzz
  @handler.check_user_access(need_privileged_access=False)
  def get(self):
    """Handle a get request."""
    data_bundles = list(data_types.DataBundle.query().order(
        data_types.DataBundle.name))

    template_values = {
        'corpora': data_bundles,
        'fieldValues': {
            'csrfToken': form.generate_csrf_token(),
            'createUrl': '/corpora/create',
            'deleteUrl': '/corpora/delete',
        },
    }
    return self.render('corpora.html', template_values)


class CreateHandler(base_handler.Handler):
  """Create a corpus."""

  @handler.post(handler.FORM, handler.HTML)
  @handler.check_user_access(need_privileged_access=True)
  @handler.require_csrf_token
  def post(self):
    """Handle a post request."""
    name = request.get('name')
    if not name:
      raise helpers.EarlyExitException('Please give this corpus a name!', 400)

    if not data_types.DataBundle.VALID_NAME_REGEX.match(name):
      raise helpers.EarlyExitException(
          'Name can only contain letters, numbers, dashes and underscores.',
          400)

    user_email = helpers.get_user_email()
    bucket_name = data_handler.get_data_bundle_bucket_name(name)
    bucket_url = data_handler.get_data_bundle_bucket_url(name)
    is_local = not request.get('nfs', False)

    if not data_handler.create_data_bundle_bucket_and_iams(name, [user_email]):
      raise helpers.EarlyExitException(
          'Failed to create bucket %s.' % bucket_name, 400)

    data_bundle = data_types.DataBundle.query(
        data_types.DataBundle.name == name).get()

    if not data_bundle:
      data_bundle = data_types.DataBundle()
    data_bundle.name = name
    data_bundle.bucket_name = bucket_name
    data_bundle.is_local = is_local
    data_bundle.put()

    template_values = {
        'title':
            'Success',
        'message': (
            'Upload data to the corpus using: '
            'gsutil -d -m rsync -r <local_corpus_directory> %s' % bucket_url),
    }
    return self.render('message.html', template_values)


class DeleteHandler(base_handler.Handler):
  """Delete a corpus."""

  @handler.post(handler.FORM, handler.HTML)
  @handler.check_user_access(need_privileged_access=True)
  @handler.require_csrf_token
  def post(self):
    """Handle a post request."""
    key = helpers.get_integer_key(request)

    data_bundle = ndb.Key(data_types.DataBundle, key).get()
    if not data_bundle:
      raise helpers.EarlyExitException('Corpus not found', 400)

    affected_fuzzers = data_types.Fuzzer.query(
        data_types.Fuzzer.data_bundle_name == data_bundle.name)
    for fuzzer in affected_fuzzers:
      fuzzer.data_bundle_name = None
      fuzzer.put()

    data_bundle.key.delete()

    template_values = {
        'title':
            'Success',
        'message': ('Corpus %s is successfully deleted. '
                    'Redirecting back to corpora page...') % data_bundle.name,
        'redirect_url':
            '/corpora',
    }
    return self.render('message.html', template_values)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Generate inputs using ml rnn model."""

import argparse
import math
import os
import sys
import time

import numpy as np
import tensorflow as tf

from clusterfuzz._internal.bot.fuzzers.ml.rnn import constants
from clusterfuzz._internal.bot.fuzzers.ml.rnn import utils

# Reset batch_size for generation: generate multiple inputs in each run.
BATCH_SIZE = 100

# Pick one byte from topn bytes with highest probabilities.
# It's recommended to use 10 for intermediate checkpoint models, since the
# model is less accurate; use a smaller value for fully trained models.
# The larger it is set, the more randomness we give.
TOPN = 4

# The upper limit for bytes generated in each round. Having this limit
# guarantees that some units can be generated within 10 minutes.
UPPER_LENGTH_LIMIT = 10000

# The lower limit for bytes generated in each round. This is to avoid
# duplicate generation for small units.
LOWER_LENGTH_LIMIT = 4


def main(args):
  """Main function to generate inputs.

  Args:
    args: Parsed arguments.

  Returns:
    Execution status defined by `constants.ExitCode`.
  """
  # Validate required paths.
  if not validate_paths(args):
    return constants.ExitCode.INVALID_PATH

  # Extract paths.
  input_dir = args.input_dir
  output_dir = args.output_dir
  model_path = args.model_path

  # Extract model parameters.
  count = args.count
  hidden_state_size = args.hidden_state_size
  hidden_layer_size = args.hidden_layer_size

  # Use timestamp as part of identifier for each testcase generated.
  timestamp = str(math.trunc(time.time()))

  print('\nusing model {} to generate {} inputs...'.format(model_path, count))

  # Restore the RNN model by building it and loading the weights.
  model = utils.build_model(hidden_layer_size * hidden_state_size,
                            constants.DROPOUT_PKEEP, constants.BATCH_SIZE,
                            False)
  try:
    model.load_weights(model_path)
  except ValueError:
    print('Incompatible model parameters.', file=sys.stderr)
    return constants.ExitCode.TENSORFLOW_ERROR

  model.build(tf.TensorShape([constants.BATCH_SIZE, None]))
  model.reset_states()

  corpus_files_info = utils.get_files_info(input_dir)
  if not corpus_files_info:
    return constants.ExitCode.CORPUS_TOO_SMALL

  new_units_count = 0
  while new_units_count < count:
    # Reset hidden states each time to generate new inputs, so that
    # different rounds will not interfere.
    model.reset_states()

    # Randomly select `BATCH_SIZE` number of inputs from corpus.
    # Record their first byte and file length.
    new_files_bytes = []
    corpus_files_length = []
    for i in range(BATCH_SIZE):
      file_info = utils.random_element_from_list(corpus_files_info)
      first_byte, file_size = file_info['first_byte'], file_info['file_size']
      new_files_bytes.append([first_byte])
      corpus_files_length.append(file_size)

    # Use 1st and 3rd quartile values as lower and upper bound respectively.
    # Also make sure they are within upper and lower bounds.
    max_length = int(np.percentile(corpus_files_length, 75))
    max_length = min(max_length, UPPER_LENGTH_LIMIT)

    min_length = int(np.percentile(corpus_files_length, 25))
    min_length = max(LOWER_LENGTH_LIMIT, min_length)

    # Reset in special cases where min_length exceeds upper limit.
    if min_length >= max_length:
      min_length = LOWER_LENGTH_LIMIT

    input_bytes = np.array(new_files_bytes)

    for _ in range(max_length - 1):
      try:
        output = model(input_bytes).numpy()
      except tf.errors.InvalidArgumentError:
        print(
            ('Failed to run TensorFlow operations since '
             'model parameters do not match.'),
            file=sys.stderr)
        return constants.ExitCode.TENSORFLOW_ERROR

      for i in range(BATCH_SIZE):
        predicted_byte = utils.sample_from_probabilities(output[i], topn=TOPN)
        new_files_bytes[i].append(predicted_byte)
        input_bytes[i][0] = predicted_byte

    # Use timestamp as part of file name.
    for i in range(BATCH_SIZE):
      new_file_name = '{}_{:0>8}'.format(timestamp, new_units_count)
      new_file_path = os.path.join(output_dir, new_file_name)

      # Use existing input length if possible, but make sure it is between
      # min_length and max_length.
      new_file_length = max(min_length, min(corpus_files_length[i], max_length))
      new_file_byte_array = bytearray(new_files_bytes[i][:new_file_length])

      with open(new_file_path, 'wb') as new_file:
        new_file.write(new_file_byte_array)
      print('generate input: {}, feed byte: {}, input length: {}'.format(
          new_file_path, new_files_bytes[i][0], new_file_length))

      # Have we got enough inputs?
      new_units_count += 1
      if new_units_count >= count:
        break

  print('Done.')
  return constants.ExitCode.SUCCESS


def validate_paths(args):
  """Validate required paths.

  Args:
    args: Parsed arguments.

  Returns:
    True if all paths are valid, False otherwise.
  """
  if not os.path.exists(args.input_dir):
    print(
        'Input directory {} does not exist'.format(args.input_dir),
        file=sys.stderr)
    return False

  if not utils.validate_model_path(args.model_path):
    print('Model {} does not exist'.format(args.model_path), file=sys.stderr)
    return False

  if not os.path.exists(args.output_dir):
    os.mkdir(args.output_dir)

  return True


def parse_args():
  """Parse command line arguments.

  Returns:
    Parsed arguement object.
  """
  parser = argparse.ArgumentParser(
      'Generating testcases using the model trained with train.py script.')

  parser.add_argument('--input-dir', help='Input folder path', required=True)
  parser.add_argument('--output-dir', help='Output folder path', required=True)
  parser.add_argument(
      '--model-path', help='Path to trained model', required=True)
  parser.add_argument(
      '--count',
      help='Number of similar inputs to generate',
      type=int,
      required=True)

  # Optional arguments: model parameters.
  # Warning: parameter values must match the model specified above.
  parser.add_argument(
      '--hidden-state-size',
      help='Hidden state size of LSTM cell (must match model)',
      type=int,
      default=constants.HIDDEN_STATE_SIZE)
  parser.add_argument(
      '--hidden-layer-size',
      help='Hidden layer size of LSTM model (must match model)',
      type=int,
      default=constants.HIDDEN_LAYER_SIZE)

  return parser.parse_args()


if __name__ == '__main__':
  parsed_args = parse_args()
  sys.exit(main(parsed_args))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Display information for revision ranges."""

from flask import request

from clusterfuzz._internal.build_management import revisions
from handlers import base_handler
from libs import handler
from libs import helpers


class Handler(base_handler.Handler):
  """Information on a revision range."""

  @handler.get(handler.HTML)
  def get(self):
    """GET handler."""
    job_type = request.get('job')
    revision = request.get('revision')
    revision_range = request.get('range')

    if revision:
      if not revision.isdigit():
        raise helpers.EarlyExitException('Revision is not an integer.', 400)
      start_revision = end_revision = revision
    elif revision_range:
      try:
        start_revision, end_revision = revision_range.split(':')
      except:
        raise helpers.EarlyExitException('Bad revision range.', 400)

      if not start_revision.isdigit():
        raise helpers.EarlyExitException('Start revision is not an integer.',
                                         400)
      if not end_revision.isdigit():
        raise helpers.EarlyExitException('End revision is not an integer.', 400)
    else:
      raise helpers.EarlyExitException('No revision specified.', 400)

    component_revisions_list = revisions.get_component_range_list(
        start_revision, end_revision, job_type)
    if not component_revisions_list:
      raise helpers.EarlyExitException('Failed to get component revisions.',
                                       400)

    return self.render(
        'revisions-info.html',
        {'info': {
            'componentRevisionsList': component_revisions_list
        }})



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for setup."""

import os
import unittest

from pyfakefs import fake_filesystem_unittest

from clusterfuzz._internal.bot.tasks import setup
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import test_utils


class IsDirectoryOnNfsTest(unittest.TestCase):
  """Tests for the is_directory_on_nfs function."""

  def setUp(self):
    environment.set_value('NFS_ROOT', '/nfs')

  def tearDown(self):
    environment.remove_key('NFS_ROOT')

  def test_is_directory_on_nfs_without_nfs(self):
    """Test is_directory_on_nfs without nfs."""
    environment.remove_key('NFS_ROOT')
    self.assertFalse(setup.is_directory_on_nfs('/nfs/dir1'))

  def test_is_directory_on_nfs_with_nfs_and_data_bundle_on_nfs(self):
    """Test is_directory_on_nfs with nfs and data bundle on nfs."""
    self.assertTrue(setup.is_directory_on_nfs('/nfs/dir1'))

  def test_is_directory_on_nfs_with_nfs_and_data_bundle_on_local(self):
    """Test is_directory_on_nfs with nfs and data bundle on local."""
    self.assertFalse(setup.is_directory_on_nfs('/tmp/dir1'))


# pylint: disable=protected-access
@test_utils.with_cloud_emulators('datastore')
class GetApplicationArgumentsTest(unittest.TestCase):
  """Tests _get_application_arguments."""

  def setUp(self):
    helpers.patch_environ(self)

    data_types.Job(
        name='linux_asan_chrome',
        environment_string=('APP_ARGS = --orig-arg1 --orig-arg2')).put()
    data_types.Job(
        name='linux_msan_chrome_variant',
        environment_string=(
            'APP_ARGS = --arg1 --arg2 --arg3="--flag1 --flag2"')).put()

    data_types.Job(name='libfuzzer_asan_chrome', environment_string=('')).put()
    data_types.Job(
        name='libfuzzer_msan_chrome_variant', environment_string=('')).put()
    data_types.Job(
        name='afl_asan_chrome_variant', environment_string=('')).put()

    self.testcase = test_utils.create_generic_testcase()

  def test_no_minimized_arguments(self):
    """Test that None is returned when minimized arguments is not set."""
    self.testcase.minimized_arguments = ''
    self.testcase.job_type = 'linux_asan_chrome'
    self.testcase.put()

    self.assertEqual(
        None,
        setup._get_application_arguments(self.testcase, 'linux_asan_chrome',
                                         'minimize'))
    self.assertEqual(
        None,
        setup._get_application_arguments(
            self.testcase, 'linux_msan_chrome_variant', 'variant'))

  def test_minimized_arguments_for_non_variant_task(self):
    """Test that minimized arguments are returned for non-variant tasks."""
    self.testcase.minimized_arguments = '--orig-arg2'
    self.testcase.job_type = 'linux_asan_chrome'
    self.testcase.put()

    self.assertEqual(
        '--orig-arg2',
        setup._get_application_arguments(self.testcase, 'linux_asan_chrome',
                                         'minimize'))

  def test_no_unique_minimized_arguments_for_variant_task(self):
    """Test that only APP_ARGS is returned if minimized arguments have no
    unique arguments, for variant task."""
    self.testcase.minimized_arguments = '--arg2'
    self.testcase.job_type = 'linux_asan_chrome'
    self.testcase.put()

    self.assertEqual(
        '--arg1 --arg2 --arg3="--flag1 --flag2"',
        setup._get_application_arguments(
            self.testcase, 'linux_msan_chrome_variant', 'variant'))

  def test_some_duplicate_minimized_arguments_for_variant_task(self):
    """Test that both minimized arguments and APP_ARGS are returned with
    duplicate args stripped from minimized arguments for variant task."""
    self.testcase.minimized_arguments = '--arg3="--flag1 --flag2" --arg4'
    self.testcase.job_type = 'linux_asan_chrome'
    self.testcase.put()

    self.assertEqual(
        '--arg4 --arg1 --arg2 --arg3="--flag1 --flag2"',
        setup._get_application_arguments(
            self.testcase, 'linux_msan_chrome_variant', 'variant'))

  def test_unique_minimized_arguments_for_variant_task(self):
    """Test that both minimized arguments and APP_ARGS are returned when they
    don't have common args for variant task."""
    self.testcase.minimized_arguments = '--arg5'
    self.testcase.job_type = 'linux_asan_chrome'
    self.testcase.put()

    self.assertEqual(
        '--arg5 --arg1 --arg2 --arg3="--flag1 --flag2"',
        setup._get_application_arguments(
            self.testcase, 'linux_msan_chrome_variant', 'variant'))

  def test_no_job_app_args_for_variant_task(self):
    """Test that only minimized arguments is returned when APP_ARGS is not set
    in job definition."""
    self.testcase.minimized_arguments = '--arg5'
    self.testcase.job_type = 'libfuzzer_asan_chrome'
    self.testcase.put()

    self.assertEqual(
        '--arg5',
        setup._get_application_arguments(
            self.testcase, 'libfuzzer_msan_chrome_variant', 'variant'))

  def test_afl_job_for_variant_task(self):
    """Test that we use a different argument list if this is an afl variant
    task."""
    self.testcase.minimized_arguments = '--arg5'
    self.testcase.job_type = 'libfuzzer_asan_chrome'
    self.testcase.put()

    self.assertEqual(
        '%TESTCASE%',
        setup._get_application_arguments(self.testcase,
                                         'afl_asan_chrome_variant', 'variant'))


# pylint: disable=protected-access
class ClearOldDataBundlesIfNeededTest(fake_filesystem_unittest.TestCase):
  """Tests _clear_old_data_bundles_if_needed."""

  def setUp(self):
    test_utils.set_up_pyfakefs(self)
    helpers.patch_environ(self)

    self.data_bundles_dir = '/data-bundles'
    os.mkdir(self.data_bundles_dir)
    environment.set_value('DATA_BUNDLES_DIR', self.data_bundles_dir)

  def test_evict(self):
    """Tests that eviction works when more than certain number of bundles."""
    for i in range(1, 15):
      os.mkdir(os.path.join(self.data_bundles_dir, str(i)))

    setup._clear_old_data_bundles_if_needed()
    self.assertEqual([str(i) for i in range(5, 15)],
                     sorted(os.listdir(self.data_bundles_dir), key=int))

  def test_no_evict(self):
    """Tests that no eviction is required when less than certain number of
    bundles."""
    for i in range(1, 5):
      os.mkdir(os.path.join(self.data_bundles_dir, str(i)))

    setup._clear_old_data_bundles_if_needed()
    self.assertEqual([str(i) for i in range(1, 5)],
                     sorted(os.listdir(self.data_bundles_dir), key=int))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for process_handler."""

import unittest

import mock

from clusterfuzz._internal.system import process_handler
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers


class MockProcess(object):
  """Mock process."""

  def __init__(self, pid, name, cmdline):
    self._info = {
        'name': name,
        'pid': pid,
        'cmdline': cmdline,
    }

  def as_dict(self, attrs):  # pylint: disable=unused-argument
    return self._info


class TerminateProcessesMatchingNameTest(unittest.TestCase):
  """Tests terminate_processes_matching_names."""

  def setUp(self):
    test_helpers.patch(self, [
        'clusterfuzz._internal.system.process_handler.terminate_process',
        'psutil.process_iter',
    ])
    self.mock.process_iter.return_value = [
        MockProcess(1, 'process_1', ['/a/b/c', '-f1']),
        MockProcess(2, 'process_2', ['/d'])
    ]

  def test_process_1_with_terminate(self):
    process_handler.terminate_processes_matching_names('process_1')
    self.mock.terminate_process.assert_has_calls([
        mock.call(1, False),
    ])

  def test_process_1_with_kill(self):
    process_handler.terminate_processes_matching_names('process_1', kill=True)
    self.mock.terminate_process.assert_has_calls([
        mock.call(1, True),
    ])

  def test_process_2_with_terminate(self):
    process_handler.terminate_processes_matching_names('process_2')
    self.mock.terminate_process.assert_has_calls([
        mock.call(2, False),
    ])

  def test_process_2_with_kill(self):
    process_handler.terminate_processes_matching_names('process_2', kill=True)
    self.mock.terminate_process.assert_has_calls([
        mock.call(2, True),
    ])

  def test_no_process_terminate(self):
    process_handler.terminate_processes_matching_names('not_exist')
    self.assertEqual(0, self.mock.terminate_process.call_count)

  def test_no_process_kill(self):
    process_handler.terminate_processes_matching_names('not_exist', kill=True)
    self.assertEqual(0, self.mock.terminate_process.call_count)

  def test_no_process_terminate_with_partial_match(self):
    process_handler.terminate_processes_matching_names('process_')
    self.assertEqual(0, self.mock.terminate_process.call_count)

  def test_no_process_kill_with_partial_match(self):
    process_handler.terminate_processes_matching_names('process_', kill=True)
    self.assertEqual(0, self.mock.terminate_process.call_count)

  def test_process_1_no_terminate_with_wrong_case(self):
    process_handler.terminate_processes_matching_names('Process_1')
    self.assertEqual(0, self.mock.terminate_process.call_count)

  def test_process_1_no_kill_with_wrong_case(self):
    process_handler.terminate_processes_matching_names('Process_1', kill=True)
    self.assertEqual(0, self.mock.terminate_process.call_count)


class TerminateProcessesMatchingPathTest(unittest.TestCase):
  """Tests terminate_processes_matching_names."""

  def setUp(self):
    test_helpers.patch(self, [
        'clusterfuzz._internal.system.process_handler.terminate_process',
        'psutil.process_iter',
    ])
    self.mock.process_iter.return_value = [
        MockProcess(1, 'process_1', ['/a/b/c', '-f1']),
        MockProcess(2, 'process_2', ['/d'])
    ]

  def test_process_1_with_terminate(self):
    process_handler.terminate_processes_matching_cmd_line('c -f1')
    self.mock.terminate_process.assert_has_calls([
        mock.call(1, False),
    ])

  def test_process_1_with_kill(self):
    process_handler.terminate_processes_matching_cmd_line('/a/b/c', kill=True)
    self.mock.terminate_process.assert_has_calls([
        mock.call(1, True),
    ])

  def test_process_2_with_terminate(self):
    process_handler.terminate_processes_matching_cmd_line('/d')
    self.mock.terminate_process.assert_has_calls([
        mock.call(2, False),
    ])

  def test_process_2_with_kill(self):
    process_handler.terminate_processes_matching_cmd_line('/d', kill=True)
    self.mock.terminate_process.assert_has_calls([
        mock.call(2, True),
    ])

  def test_no_process_terminate(self):
    process_handler.terminate_processes_matching_cmd_line('not_exist')
    self.assertEqual(0, self.mock.terminate_process.call_count)

  def test_no_process_kill(self):
    process_handler.terminate_processes_matching_cmd_line(
        'not_exist', kill=True)
    self.assertEqual(0, self.mock.terminate_process.call_count)

  def test_process_1_no_terminate_with_wrong_case(self):
    process_handler.terminate_processes_matching_cmd_line('/a/b/C')
    self.assertEqual(0, self.mock.terminate_process.call_count)

  def test_process_1_no_kill_with_wrong_case(self):
    process_handler.terminate_processes_matching_cmd_line('/a/b/C', kill=True)
    self.assertEqual(0, self.mock.terminate_process.call_count)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for process."""
# pylint: disable=unused-argument

import queue
import time
import unittest

import mock

from clusterfuzz._internal.system import new_process
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import test_utils


class Signal(object):
  TERM = 0
  KILL = 1


def mock_kill(self):
  self._popen.kill()  # pylint: disable=protected-access


def mock_popen_factory(execute_time,
                       output,
                       sigterm_handler_time,
                       return_code=0):
  """Create a mock popen."""

  class MockPopen(object):
    """Mock subprocess.Popen."""
    received_signals = []

    def __init__(self, *args, **kwargs):
      """Inits the MockPopen."""
      self.start_time = time.time()
      self.signal_queue = queue.Queue()

    def poll(self):
      """Mock subprocess.Popen.poll."""
      if time.time() >= self.start_time + execute_time:
        return return_code

      return None

    def kill(self):
      """Mock subprocess.Popen.kill."""
      self.signal_queue.put(Signal.KILL)

    def terminate(self):
      """Mock subprocess.Popen.terminate."""
      self.signal_queue.put(Signal.TERM)

    def communicate(self, input_data=None):
      """Mock subprocess.Popen.communicate."""
      for i in range(2):
        timeout = execute_time if i == 0 else sigterm_handler_time
        try:
          received_signal = self.signal_queue.get(block=True, timeout=timeout)
        except queue.Empty:
          continue

        self.received_signals.append((received_signal,
                                      time.time() - self.start_time))
        if received_signal == Signal.KILL:
          break

      return output, None

  return MockPopen


class PosixProcessTest(unittest.TestCase):
  """Posix Process tests."""

  # Allowed error for time calculations.
  TIME_ERROR = 0.3

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.system.environment.platform',
        'clusterfuzz._internal.system.new_process.ChildProcess.kill'
    ])
    self.mock.platform.return_value = 'LINUX'
    self.mock.kill.side_effect = mock_kill

  def test_basic(self):
    """Tests basic command line."""
    runner = new_process.ProcessRunner('/test/path')
    self.assertEqual(runner.get_command(), ['/test/path'])

  def test_default_args(self):
    """Tests basic command line with default arguments."""
    runner = new_process.ProcessRunner(
        '/test/path', default_args=['-arg1', '-arg2'])
    self.assertEqual(runner.get_command(), ['/test/path', '-arg1', '-arg2'])

  def test_additional_args(self):
    """Tests basic command line with default arguments and additional
    arguments."""
    runner = new_process.ProcessRunner(
        '/test/path', default_args=['-arg1', '-arg2'])
    self.assertEqual(
        runner.get_command(additional_args=['-arg3', '-arg4']),
        ['/test/path', '-arg1', '-arg2', '-arg3', '-arg4'])

  def test_results_no_timeout(self):
    """Test process execution results."""
    with mock.patch('subprocess.Popen', mock_popen_factory(
        1.0, 'output', 0.0, 0)):
      runner = new_process.ProcessRunner(
          '/test/path', default_args=['-arg1', '-arg2'])
      results = runner.run_and_wait()
      self.assertEqual(results.command, ['/test/path', '-arg1', '-arg2'])
      self.assertEqual(results.return_code, 0)
      self.assertEqual(results.output, 'output')
      self.assertLess(abs(results.time_executed - 1.0), self.TIME_ERROR)
      self.assertFalse(results.timed_out)

  def test_results_timeout(self):
    """Test process execution results with timeout."""
    with mock.patch('subprocess.Popen', mock_popen_factory(
        1.0, 'output', 0.0, 0)):
      runner = new_process.ProcessRunner(
          '/test/path', default_args=['-arg1', '-arg2'])
      results = runner.run_and_wait(timeout=0.5)
      self.assertEqual(results.command, ['/test/path', '-arg1', '-arg2'])
      self.assertEqual(results.return_code, None)
      self.assertEqual(results.output, 'output')
      self.assertLess(abs(results.time_executed - 0.5), self.TIME_ERROR)
      self.assertTrue(results.timed_out)

  def test_timeout(self):
    """Tests timeout signals."""
    with mock.patch('subprocess.Popen', mock_popen_factory(1.0, '',
                                                           0.0)) as mock_popen:
      runner = new_process.ProcessRunner('/test/path')
      runner.run_and_wait(timeout=0.5)

      # Single signal (SIGKILL) should arrive in 0.5 seconds.
      self.assertEqual(len(mock_popen.received_signals), 1)
      self.assertLess(
          abs(mock_popen.received_signals[0][1] - 0.5), self.TIME_ERROR)
      self.assertEqual(mock_popen.received_signals[0][0], Signal.KILL)

  def test_no_timeout(self):
    """Tests process exiting before timeout."""
    with mock.patch('subprocess.Popen', mock_popen_factory(0.5, '',
                                                           0.0)) as mock_popen:
      runner = new_process.ProcessRunner('/test/path')
      runner.run_and_wait(timeout=5.0)

      # No signals should be sent.
      self.assertEqual(len(mock_popen.received_signals), 0)

  def test_terminate_before_kill_no_sigterm_timeout(self):
    """Tests process sigterm handler completing before terminate_wait_time."""
    with mock.patch('subprocess.Popen', mock_popen_factory(1.0, '',
                                                           0.5)) as mock_popen:
      runner = new_process.ProcessRunner('/test/path')
      runner.run_and_wait(
          timeout=0.5, terminate_before_kill=True, terminate_wait_time=1.0)

      # Single signal (SIGTERM) in 0.5 seconds.
      self.assertEqual(len(mock_popen.received_signals), 1)
      self.assertLess(
          abs(mock_popen.received_signals[0][1] - 0.5), self.TIME_ERROR)
      self.assertEqual(mock_popen.received_signals[0][0], Signal.TERM)

  def test_terminate_before_kill_timeout(self):
    """Tests process sigterm handler timing out."""
    with mock.patch('subprocess.Popen', mock_popen_factory(1.0, '',
                                                           1.0)) as mock_popen:
      runner = new_process.ProcessRunner('/test/path')
      runner.run_and_wait(
          timeout=0.5, terminate_before_kill=True, terminate_wait_time=0.5)

      # First signal (SIGTERM) should arrive in 0.5 seconds.
      self.assertEqual(len(mock_popen.received_signals), 2)
      self.assertLess(
          abs(mock_popen.received_signals[0][1] - 0.5), self.TIME_ERROR)
      self.assertEqual(mock_popen.received_signals[0][0], Signal.TERM)

      # First signal (SIGKILL) should arrive in 1.0 (0.5 + 0.5) seconds.
      self.assertLess(
          abs(mock_popen.received_signals[1][1] - 1.0), self.TIME_ERROR)
      self.assertEqual(mock_popen.received_signals[1][0], Signal.KILL)


class WindowsProcessTest(unittest.TestCase):
  """Windows Process tests."""

  # Allowed error for time calculations.
  TIME_ERROR = 0.3

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.system.environment.platform',
        'clusterfuzz._internal.system.new_process.ChildProcess.kill'
    ])
    self.mock.kill.side_effect = mock_kill
    self.mock.platform.return_value = 'WINDOWS'

  def test_terminate_before_kill_timeout(self):
    """Tests process kill handler called on timeout."""
    with mock.patch('subprocess.Popen', mock_popen_factory(1.0, '',
                                                           1.0)) as mock_popen:
      runner = new_process.ProcessRunner('/test/path')
      runner.run_and_wait(
          timeout=0.5, terminate_before_kill=True, terminate_wait_time=0.5)

      # Single signal (SIGKILL) should arrive in 0.5 seconds.
      self.assertEqual(len(mock_popen.received_signals), 1)
      self.assertLess(
          abs(mock_popen.received_signals[0][1] - 0.5), self.TIME_ERROR)
      self.assertEqual(mock_popen.received_signals[0][0], Signal.KILL)


@test_utils.integration
class MaxStdoutLenTest(unittest.TestCase):
  """Test max_stdout_len."""

  def test_over_limit(self):
    """Test stdout over limit."""
    runner = new_process.ProcessRunner('python')
    result = runner.run_and_wait(
        ['-c', 'print("A" + "B"*499 + "C"*499 + "D")'], max_stdout_len=64)
    self.assertEqual(
        b'A' + b'B' * 31 + b'\n...truncated 937 bytes...\n' + b'C' * 30 + b'D' +
        b'\n', result.output)

  def test_under_limit(self):
    """Test stdout under limit."""
    runner = new_process.ProcessRunner('python')
    result = runner.run_and_wait(['-c', 'print("A"*62)'], max_stdout_len=64)
    self.assertEqual(b'A' * 62 + b'\n', result.output)


if __name__ == '__main__':
  unittest.main()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handler for updating issue."""

from flask import request

from clusterfuzz._internal.datastore import data_handler
from handlers import base_handler
from handlers.testcase_detail import show
from libs import handler
from libs import helpers
from libs.issue_management import issue_filer
from libs.issue_management import issue_tracker_policy


class Handler(base_handler.Handler):
  """Handler that updates an issue."""

  @staticmethod
  def update_issue(testcase, issue_id, needs_summary_update):
    """Associate (or update) an existing issue with the testcase."""
    issue_id = helpers.cast(issue_id, int,
                            'Issue ID (%s) is not a number!' % issue_id)
    issue_tracker = helpers.get_issue_tracker_for_testcase(testcase)

    issue = helpers.get_or_exit(lambda: issue_tracker.get_issue(issue_id),
                                'Issue (id=%d) is not found!' % issue_id,
                                'Failed to get the issue (id=%s).' % issue_id,
                                Exception)

    if not issue.is_open:
      raise helpers.EarlyExitException(
          ('The issue (%d) is already closed and further updates are not'
           ' allowed. Please file a new issue instead!') % issue_id, 400)

    if not testcase.is_crash():
      raise helpers.EarlyExitException(
          'This is not a crash testcase, so issue update is not applicable.',
          400)

    issue_comment = data_handler.get_issue_description(testcase,
                                                       helpers.get_user_email())
    if needs_summary_update:
      issue.title = data_handler.get_issue_summary(testcase)

    policy = issue_tracker_policy.get(issue_tracker.project)
    properties = policy.get_existing_issue_properties()
    for label in properties.labels:
      for result in issue_filer.apply_substitutions(policy, label, testcase):
        issue.labels.add(result)

    issue.save(new_comment=issue_comment)

    testcase.bug_information = str(issue_id)
    testcase.put()

    data_handler.update_group_bug(testcase.group_id)

    helpers.log('Updated issue %sd' % issue_id, helpers.MODIFY_OPERATION)

  @handler.post(handler.JSON, handler.JSON)
  @handler.require_csrf_token
  @handler.check_admin_access_if_oss_fuzz
  @handler.check_testcase_access
  def post(self, testcase):
    """Update an issue."""
    issue_id = request.get('issueId')
    needs_summary_update = request.get('needsSummaryUpdate')

    self.update_issue(testcase, issue_id, needs_summary_update)
    return self.render_json(show.get_testcase_detail(testcase))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Classes for dealing with FuzzerStats."""

import datetime
import functools
import itertools
import json
import os
import random
import re

from clusterfuzz._internal.base import memoize
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.datastore import fuzz_target_utils
from clusterfuzz._internal.google_cloud_utils import big_query
from clusterfuzz._internal.google_cloud_utils import storage
from clusterfuzz._internal.metrics import fuzzer_logs
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import shell

STATS_FILE_EXTENSION = '.stats2'

PERFORMANCE_REPORT_VIEWER_PATH = '/performance-report/{fuzzer}/{job}/{date}'

JOB_RUN_SCHEMA = {
    'fields': [{
        'name': 'testcases_executed',
        'type': 'INTEGER',
        'mode': 'NULLABLE'
    }, {
        'name': 'build_revision',
        'type': 'INTEGER',
        'mode': 'NULLABLE'
    }, {
        'name': 'new_crashes',
        'type': 'INTEGER',
        'mode': 'NULLABLE'
    }, {
        'name': 'job',
        'type': 'STRING',
        'mode': 'NULLABLE'
    }, {
        'name': 'timestamp',
        'type': 'FLOAT',
        'mode': 'NULLABLE'
    }, {
        'name':
            'crashes',
        'type':
            'RECORD',
        'mode':
            'REPEATED',
        'fields': [{
            'name': 'crash_type',
            'type': 'STRING',
            'mode': 'NULLABLE'
        }, {
            'name': 'is_new',
            'type': 'BOOLEAN',
            'mode': 'NULLABLE'
        }, {
            'name': 'crash_state',
            'type': 'STRING',
            'mode': 'NULLABLE'
        }, {
            'name': 'security_flag',
            'type': 'BOOLEAN',
            'mode': 'NULLABLE'
        }, {
            'name': 'count',
            'type': 'INTEGER',
            'mode': 'NULLABLE'
        }]
    }, {
        'name': 'known_crashes',
        'type': 'INTEGER',
        'mode': 'NULLABLE'
    }, {
        'name': 'fuzzer',
        'type': 'STRING',
        'mode': 'NULLABLE'
    }, {
        'name': 'kind',
        'type': 'STRING',
        'mode': 'NULLABLE'
    }]
}


class FuzzerStatsException(Exception):
  """Fuzzer stats exception."""


class BaseRun(object):
  """Base run."""

  VALID_FIELDNAME_PATTERN = re.compile(r'[a-zA-Z][a-zA-Z0-9_]*')

  def __init__(self, fuzzer, job, build_revision, timestamp):
    self._stats_data = {
        'fuzzer': fuzzer,
        'job': job,
        'build_revision': build_revision,
        'timestamp': timestamp,
    }

  def __getitem__(self, key):
    return self._stats_data.__getitem__(key)

  def __setitem__(self, key, value):
    if not re.compile(self.VALID_FIELDNAME_PATTERN):
      raise ValueError('Invalid key name.')

    return self._stats_data.__setitem__(key, value)

  def __delitem__(self, key):
    return self._stats_data.__delitem__(key)

  def __contains__(self, key):
    return self._stats_data.__contains__(key)

  def to_json(self):
    """Return JSON representation of the stats."""
    return json.dumps(self._stats_data)

  def update(self, other):
    """Update stats with a dict."""
    self._stats_data.update(other)

  @property
  def data(self):
    return self._stats_data

  @property
  def kind(self):
    return self._stats_data['kind']

  @property
  def fuzzer(self):
    return self._stats_data['fuzzer']

  @property
  def job(self):
    return self._stats_data['job']

  @property
  def build_revision(self):
    return self._stats_data['build_revision']

  @property
  def timestamp(self):
    return self._stats_data['timestamp']

  @staticmethod
  def from_json(json_data):
    """Convert json to the run."""
    try:
      data = json.loads(json_data)
    except (ValueError, TypeError):
      return None

    if not isinstance(data, dict):
      return None

    result = None
    try:
      kind = data['kind']
      if kind == 'TestcaseRun':
        result = TestcaseRun(data['fuzzer'], data['job'],
                             data['build_revision'], data['timestamp'])
      elif kind == 'JobRun':
        result = JobRun(data['fuzzer'], data['job'], data['build_revision'],
                        data['timestamp'], data['testcases_executed'],
                        data['new_crashes'], data['known_crashes'],
                        data.get('crashes'))
    except KeyError:
      return None

    if result:
      result.update(data)
    return result


class JobRun(BaseRun):
  """Represents stats for a particular job run."""

  SCHEMA = JOB_RUN_SCHEMA

  # `crashes` is a new field that will replace `new_crashes` and `old_crashes`.
  def __init__(self, fuzzer, job, build_revision, timestamp,
               number_of_testcases, new_crashes, known_crashes, crashes):
    super(JobRun, self).__init__(fuzzer, job, build_revision, timestamp)
    self._stats_data.update({
        'kind': 'JobRun',
        'testcases_executed': number_of_testcases,
        'new_crashes': new_crashes,
        'known_crashes': known_crashes,
        'crashes': crashes
    })


class TestcaseRun(BaseRun):
  """Represents stats for a particular testcase run."""

  SCHEMA = None

  def __init__(self, fuzzer, job, build_revision, timestamp):
    super(TestcaseRun, self).__init__(fuzzer, job, build_revision, timestamp)
    self._stats_data.update({
        'kind': 'TestcaseRun',
    })

    source = environment.get_value('STATS_SOURCE')
    if source:
      self._stats_data['source'] = source

  @staticmethod
  def get_stats_filename(testcase_file_path):
    """Get stats filename for the given testcase."""
    return testcase_file_path + STATS_FILE_EXTENSION

  @staticmethod
  def read_from_disk(testcase_file_path, delete=False):
    """Read the TestcaseRun for the given testcase."""
    stats_file_path = TestcaseRun.get_stats_filename(testcase_file_path)
    if not os.path.exists(stats_file_path):
      return None

    fuzzer_run = None
    with open(stats_file_path) as f:
      fuzzer_run = BaseRun.from_json(f.read())

    if delete:
      shell.remove_file(stats_file_path)

    return fuzzer_run

  @staticmethod
  def write_to_disk(testcase_run, testcase_file_path):
    """Write the given TestcaseRun for |testcase_file_path| to disk."""
    if not testcase_run:
      return

    stats_file_path = TestcaseRun.get_stats_filename(testcase_file_path)
    with open(stats_file_path, 'w') as f:
      f.write(testcase_run.to_json())


class QueryGroupBy(object):
  """GroupBy enum."""

  GROUP_BY_NONE = 0
  GROUP_BY_REVISION = 1
  GROUP_BY_DAY = 2
  GROUP_BY_TIME = 3
  GROUP_BY_JOB = 4
  GROUP_BY_FUZZER = 5


def group_by_to_field_name(group_by):
  """Convert QueryGroupBy value to its corresponding field name."""
  if group_by == QueryGroupBy.GROUP_BY_REVISION:
    return 'build_revision'

  if group_by == QueryGroupBy.GROUP_BY_DAY:
    return 'date'

  if group_by == QueryGroupBy.GROUP_BY_TIME:
    return 'time'

  if group_by == QueryGroupBy.GROUP_BY_JOB:
    return 'job'

  if group_by == QueryGroupBy.GROUP_BY_FUZZER:
    return 'fuzzer'

  return None


class BuiltinFieldData(object):
  """Represents a cell value for a builtin field."""

  def __init__(self, value, sort_key=None, link=None):
    self.value = value
    self.sort_key = sort_key
    self.link = link


class BuiltinFieldSpecifier(object):
  """Represents a builtin field."""

  def __init__(self, name, alias=None):
    self.name = name
    self.alias = alias

  def create(self, ctx=None):
    """Create the actual BuiltinField."""
    constructor = BUILTIN_FIELD_CONSTRUCTORS.get(self.name)
    if not constructor:
      return None

    return constructor(ctx)

  def field_class(self):
    """Return the class for the field."""
    constructor = BUILTIN_FIELD_CONSTRUCTORS.get(self.name)
    if not constructor:
      return None

    if isinstance(constructor, functools.partial):
      return constructor.func

    return constructor


class BuiltinField(object):
  """Base Builtin field."""

  def __init__(self, ctx=None):
    self.ctx = ctx

  def get(self, group_by, group_by_value):  # pylint: disable=unused-argument
    """Return BuiltinFieldData."""
    return None


class BuiltinFieldContext(object):
  """Context for builtin fields."""

  def __init__(self, fuzzer=None, jobs=None):
    self.fuzzer = fuzzer
    self.jobs = jobs

  def single_job_or_none(self):
    """Return the job if only 1 is specified, or None."""
    if self.jobs and len(self.jobs) == 1:
      return self.jobs[0]

    return None


class CoverageFieldContext(BuiltinFieldContext):
  """Coverage field context. Acts as a cache."""

  def __init__(self, fuzzer=None, jobs=None):
    super(CoverageFieldContext, self).__init__(fuzzer=fuzzer, jobs=jobs)

  @memoize.wrap(memoize.FifoInMemory(256))
  def get_coverage_info(self, fuzzer, date=None):
    """Return coverage info of child fuzzers."""
    if fuzzer in data_types.BUILTIN_FUZZERS:
      # Get coverage info for a job (i.e. a project).
      job = self.single_job_or_none()
      project = data_handler.get_project_name(job)
      return get_coverage_info(project, date)

    fuzz_target = data_handler.get_fuzz_target(fuzzer)
    if fuzz_target:
      fuzzer = fuzz_target.project_qualified_name()

    return get_coverage_info(fuzzer, date)


class BaseCoverageField(object):
  """Base builtin field class for coverage related fields."""

  CONTEXT_CLASS = CoverageFieldContext

  def __init__(self, ctx):
    self.ctx = ctx

  def get_coverage_info(self, group_by, group_by_value):
    """Return coverage information."""
    coverage_info = None
    if group_by == QueryGroupBy.GROUP_BY_DAY:
      # Return coverage data for the fuzzer and the day.
      coverage_info = self.ctx.get_coverage_info(self.ctx.fuzzer,
                                                 group_by_value)

    elif group_by == QueryGroupBy.GROUP_BY_FUZZER:
      # Return latest coverage data for each fuzzer.
      coverage_info = self.ctx.get_coverage_info(group_by_value)

    elif group_by == QueryGroupBy.GROUP_BY_JOB:
      # Return the latest coverage data for the fuzzer. Even though we group by
      # job here, coverage information does not differ across jobs. As of now,
      # it only depends on the fuzzer name and the date.
      coverage_info = self.ctx.get_coverage_info(self.ctx.fuzzer)

    return coverage_info


class CoverageField(BaseCoverageField):
  """Coverage field."""

  EDGE = 0
  FUNCTION = 1
  VALUE_TYPE = float

  def __init__(self, coverage_type, ctx=None):
    super(CoverageField, self).__init__(ctx)
    self.coverage_type = coverage_type

  def get(self, group_by, group_by_value):
    """Return data."""
    coverage_info = self.get_coverage_info(group_by, group_by_value)
    if not coverage_info:
      return None

    if self.coverage_type == self.EDGE:
      covered = coverage_info.edges_covered
      total = coverage_info.edges_total
    else:
      covered = coverage_info.functions_covered
      total = coverage_info.functions_total

    if covered is None or total is None:
      return None

    if not total:
      logs.log_error(
          'Invalid coverage info: total equals 0 for "%s".' % self.ctx.fuzzer)
      return BuiltinFieldData('No coverage', sort_key=0.0)

    percentage = 100.0 * float(covered) / total
    display_value = '%.2f%% (%d/%d)' % (percentage, covered, total)
    return BuiltinFieldData(display_value, sort_key=percentage)


class CorpusBackupField(BaseCoverageField):
  """Link to the latest corpus backup archive."""
  VALUE_TYPE = str

  def __init__(self, ctx=None):
    super(CorpusBackupField, self).__init__(ctx)

  def get(self, group_by, group_by_value):
    """Return data."""
    coverage_info = self.get_coverage_info(group_by, group_by_value)
    if not coverage_info:
      return None

    if not coverage_info.corpus_backup_location:
      return None

    # Google Cloud console does not support linking to a specific file, so we
    # link to the directory instead.
    corpus_backup_location = os.path.dirname(
        coverage_info.corpus_backup_location)

    display_value = 'Download'
    return BuiltinFieldData(display_value, link=corpus_backup_location)


class CorpusSizeField(BaseCoverageField):
  """Corpus size field."""

  CORPUS = 0
  QUARANTINE = 1
  VALUE_TYPE = int

  def __init__(self, corpus_type, ctx=None):
    super(CorpusSizeField, self).__init__(ctx)
    self.corpus_type = corpus_type

  def get(self, group_by, group_by_value):
    """Return data."""
    if (self.ctx.fuzzer in data_types.BUILTIN_FUZZERS and
        group_by == QueryGroupBy.GROUP_BY_DAY):
      # Explicitly return None here, as coverage_info below might exist and have
      # default corpus size of 0, which might look confusing on the stats page.
      return None

    coverage_info = self.get_coverage_info(group_by, group_by_value)
    if not coverage_info:
      return None

    if self.corpus_type == self.CORPUS:
      corpus_size_units = coverage_info.corpus_size_units
      corpus_size_bytes = coverage_info.corpus_size_bytes
      corpus_location = coverage_info.corpus_location
    else:
      corpus_size_units = coverage_info.quarantine_size_units
      corpus_size_bytes = coverage_info.quarantine_size_bytes
      corpus_location = coverage_info.quarantine_location

    # If the values aren't specified, return None to show the default '--' text.
    if corpus_size_units is None or corpus_size_bytes is None:
      return None

    display_value = '%d (%s)' % (corpus_size_units,
                                 utils.get_size_string(corpus_size_bytes))

    return BuiltinFieldData(
        display_value, sort_key=corpus_size_units, link=corpus_location)


class CoverageReportField(BaseCoverageField):
  """Coverage report field."""

  VALUE_TYPE = str

  def __init__(self, ctx=None):
    super(CoverageReportField, self).__init__(ctx)

  def get(self, group_by, group_by_value):
    """Return data."""
    coverage_info = self.get_coverage_info(group_by, group_by_value)
    if not coverage_info or not coverage_info.html_report_url:
      return None

    display_value = 'Coverage'
    return BuiltinFieldData(display_value, link=coverage_info.html_report_url)


def _logs_bucket_key_fn(func, args, kwargs):  # pylint: disable=unused-argument
  return 'fuzzer_logs_bucket:' + args[1]


class FuzzerRunLogsContext(BuiltinFieldContext):
  """Fuzzer logs context."""

  MEMCACHE_TTL = 30 * 60

  def __init__(self, fuzzer=None, jobs=None):
    super(FuzzerRunLogsContext, self).__init__(fuzzer=fuzzer, jobs=jobs)

  @memoize.wrap(memoize.FifoInMemory(1024))
  def _get_logs_bucket_from_job(self, job_type):
    """Get logs bucket from job."""
    return data_handler.get_value_from_job_definition_or_environment(
        job_type, 'FUZZ_LOGS_BUCKET')

  @memoize.wrap(memoize.Memcache(MEMCACHE_TTL, key_fn=_logs_bucket_key_fn))
  def _get_logs_bucket_from_fuzzer(self, fuzzer_name):
    """Get logs bucket from fuzzer (child fuzzers only)."""
    jobs = [
        mapping.job for mapping in fuzz_target_utils.get_fuzz_target_jobs(
            fuzz_target_name=fuzzer_name)
    ]
    if not jobs:
      return None

    # Check that the logs bucket is same for all of them.
    bucket = self._get_logs_bucket_from_job(jobs[0])
    if all(bucket == self._get_logs_bucket_from_job(job) for job in jobs[1:]):
      return bucket

    return None

  def get_logs_bucket(self, fuzzer_name=None, job_type=None):
    """Return logs bucket for the job."""
    if job_type:
      return self._get_logs_bucket_from_job(job_type)

    if fuzzer_name:
      return self._get_logs_bucket_from_fuzzer(fuzzer_name)

    return None


class FuzzerRunLogsField(BuiltinField):
  """Fuzzer logs field."""

  CONTEXT_CLASS = FuzzerRunLogsContext
  VALUE_TYPE = str

  def _get_logs_bucket_path(self, group_by, group_by_value):
    """Return logs bucket path."""
    fuzzer = self.ctx.fuzzer
    job = self.ctx.single_job_or_none()
    date = None

    if group_by == QueryGroupBy.GROUP_BY_FUZZER:
      fuzzer = group_by_value
    elif group_by == QueryGroupBy.GROUP_BY_DAY:
      if not fuzzer:
        return None
      if not job:
        # If job isn't specified, we'll ignore the date and show the link to the
        # GCS directory containing all jobs for a given fuzz target, because job
        # name comes before the date in the GCS path.
        date = None
      else:
        date = group_by_value
    elif group_by == QueryGroupBy.GROUP_BY_JOB:
      job = group_by_value
    else:
      return None

    if not fuzzer:
      # Fuzzer always needs to be specified (first component in GCS path).
      return None

    logs_bucket = self.ctx.get_logs_bucket(fuzzer_name=fuzzer, job_type=job)
    if not logs_bucket:
      return None

    return 'gs:/' + fuzzer_logs.get_logs_directory(logs_bucket, fuzzer, job,
                                                   date)

  def get(self, group_by, group_by_value):
    """Return data."""
    logs_path = self._get_logs_bucket_path(group_by, group_by_value)
    if not logs_path:
      return None

    return BuiltinFieldData('Logs', link=logs_path)


class PerformanceReportField(BuiltinField):
  """Performance report field."""

  CONTEXT_CLASS = FuzzerRunLogsContext
  VALUE_TYPE = str

  def _get_performance_report_path(self, group_by, group_by_value):
    """Return performance analysis report path."""
    fuzzer = self.ctx.fuzzer
    job = self.ctx.single_job_or_none()
    date = 'latest'

    if group_by == QueryGroupBy.GROUP_BY_FUZZER:
      fuzzer = group_by_value
    elif group_by == QueryGroupBy.GROUP_BY_JOB:
      job = group_by_value
    elif group_by == QueryGroupBy.GROUP_BY_DAY:
      date = group_by_value
    else:
      return None

    if not fuzzer or not job:
      return None

    return PERFORMANCE_REPORT_VIEWER_PATH.format(
        fuzzer=fuzzer, job=job, date=date)

  def get(self, group_by, group_by_value):
    """Return data."""
    report_path = self._get_performance_report_path(group_by, group_by_value)
    if not report_path:
      return None

    return BuiltinFieldData('Performance', link=report_path)


class QueryField(object):
  """Represents a query field."""

  def __init__(self,
               table_alias,
               field_name,
               aggregate_function,
               select_alias=None):
    self.table_alias = table_alias
    self.name = field_name
    self.aggregate_function = aggregate_function
    self.select_alias = select_alias or field_name

  def is_custom(self):
    """Return true if this field uses complex query. This field won't appear
      in the SELECT's fields automatically. We will need to define how to get
      the data."""
    return (self.aggregate_function and
            self.aggregate_function.lower() == 'custom')

  def __str__(self):
    if self.aggregate_function:
      result = '%s(%s.%s)' % (self.aggregate_function, self.table_alias,
                              self.name)
    else:
      result = '%s.%s' % (self.table_alias, self.name)

    if self.select_alias:
      result += ' as ' + self.select_alias

    return result


class Query(object):
  """Represents a stats query."""

  def _ensure_valid_name(self, name, regex):
    """Ensure that the given name is valid for fuzzer/jobs."""
    if name and not regex.match(name):
      raise FuzzerStatsException('Invalid fuzzer or job name.')

  def __init__(self, fuzzer_name, job_types, query_fields, group_by, date_start,
               date_end, base_table, alias):
    assert group_by is not None

    self._ensure_valid_name(fuzzer_name, data_types.Fuzzer.VALID_NAME_REGEX)

    if job_types:
      for job_type in job_types:
        self._ensure_valid_name(job_type, data_types.Job.VALID_NAME_REGEX)

    self.fuzzer_name = fuzzer_name
    self.job_types = job_types
    self.query_fields = query_fields
    self.group_by = group_by
    self.date_start = date_start
    self.date_end = date_end
    self.base_table = base_table
    self.alias = alias

    self.fuzzer_or_engine_name = get_fuzzer_or_engine_name(fuzzer_name)

  def _group_by_select(self):
    """Return a group by field."""
    if self.group_by == QueryGroupBy.GROUP_BY_DAY:
      return ('TIMESTAMP_TRUNC(TIMESTAMP_SECONDS(CAST('
              'timestamp AS INT64)), DAY, "UTC") as date')

    if self.group_by == QueryGroupBy.GROUP_BY_TIME:
      return 'TIMESTAMP_SECONDS(CAST(timestamp AS INT64)) as time'

    return group_by_to_field_name(self.group_by)

  def _group_by(self):
    """Return the group by part of the query."""
    return group_by_to_field_name(self.group_by)

  def _select_fields(self):
    """Return fields for the query."""
    group_by_select = self._group_by_select()
    fields = [group_by_select] if group_by_select else []

    for field in self.query_fields:
      if field.is_custom():
        continue
      if field.aggregate_function:
        fields.append('%s(%s) as %s' % (field.aggregate_function, field.name,
                                        field.select_alias))
      else:
        fields.append('%s as %s' % (field.name, field.select_alias))

    return ', '.join(fields)

  def _table_name(self):
    """Return the table name for the query."""
    app_id = utils.get_application_id()

    dataset = dataset_name(self.fuzzer_or_engine_name)

    return '`%s`.%s.%s' % (app_id, dataset, self.base_table)

  def _where(self):
    """Return the where part of the query."""
    result = []
    result.extend(self._partition_selector())
    result.extend(self._job_and_fuzzer_selector())

    result = ' AND '.join(result)
    if result:
      return 'WHERE ' + result

    return ''

  def _job_and_fuzzer_selector(self):
    """Return the job filter condition."""
    result = []
    if self.job_types:
      result.append('(%s)' % ' OR '.join(
          ['job = \'%s\'' % job_type for job_type in self.job_types]))

    if self.fuzzer_name != self.fuzzer_or_engine_name:
      result.append('fuzzer = \'%s\'' % self.fuzzer_name)

    return result

  def _partition_selector(self):
    """Return the partition filter condition."""
    result = ('(_PARTITIONTIME BETWEEN TIMESTAMP_SECONDS(%d) '
              'AND TIMESTAMP_SECONDS(%d))')

    return [
        result % (int(utils.utc_date_to_timestamp(self.date_start)),
                  int(utils.utc_date_to_timestamp(self.date_end)))
    ]

  def build(self):
    """Return query."""
    query_parts = [
        'SELECT',
        self._select_fields(),
        'FROM',
        self._table_name(),
        self._where(),
    ]

    if self._group_by():
      query_parts += ['GROUP BY', self._group_by()]

    return ' '.join(query_parts)


class TestcaseQuery(Query):
  """The query class for TestcaseRun Query."""

  ALIAS = 't'

  def __init__(self, fuzzer_name, job_types, query_fields, group_by, date_start,
               date_end):
    super(TestcaseQuery, self).__init__(
        fuzzer_name=fuzzer_name,
        job_types=job_types,
        query_fields=query_fields,
        group_by=group_by,
        date_start=date_start,
        date_end=date_end,
        base_table='TestcaseRun',
        alias=TestcaseQuery.ALIAS)


class JobQuery(Query):
  """The query class for JobRun Query."""

  DEFAULT_FIELDS = """
    sum(j.testcases_executed) as testcases_executed,
    custom(j.total_crashes) as total_crashes,
    custom(j.new_crashes) as new_crashes,
    custom(j.known_crashes) as known_crashes
  """
  SQL = """
    WITH
      JobRunWithConcatedCrashes AS (
        SELECT
          {select_fields},
          ARRAY_CONCAT_AGG(crashes) AS crashes
        FROM
          {table_name}
        {where}
        GROUP BY
          {group_by}
      ),
      JobRunWithUniqueCrashes AS (
        SELECT
          * EXCEPT(crashes),
          ARRAY(
            SELECT AS STRUCT
              crash.crash_type,
              crash.crash_state,
              crash.security_flag,
              SUM(count) AS count,
              MAX(crash.is_new) AS is_new
            FROM
              UNNEST(crashes) AS crash
            GROUP BY
              crash.crash_type,
              crash.crash_state,
              crash.security_flag
          ) AS crashes
        FROM
          JobRunWithConcatedCrashes
      ),
      JobRunWithSummary AS (
        SELECT
          * EXCEPT(crashes),
          (
            SELECT AS STRUCT
              IFNULL(SUM(crash.count), 0) AS total,
              COUNTIF(crash.is_new) AS unique_new,
              COUNT(crash) AS unique
            FROM
              UNNEST(crashes) AS crash
          ) AS crash_count
        FROM
          JobRunWithUniqueCrashes
      )

    SELECT
      * EXCEPT(crash_count),
      crash_count.total AS total_crashes,
      crash_count.unique_new AS new_crashes,
      (crash_count.unique - crash_count.unique_new) AS known_crashes
    FROM
      JobRunWithSummary
  """
  ALIAS = 'j'

  def __init__(self, fuzzer_name, job_types, query_fields, group_by, date_start,
               date_end):

    super(JobQuery, self).__init__(
        fuzzer_name=fuzzer_name,
        job_types=job_types,
        query_fields=query_fields,
        group_by=group_by,
        date_start=date_start,
        date_end=date_end,
        base_table='JobRun',
        alias=JobQuery.ALIAS)

  def build(self):
    """Return query."""
    sql = JobQuery.SQL.format(
        table_name=self._table_name(),
        select_fields=self._select_fields(),
        group_by=self._group_by(),
        where=self._where())

    return sql


class TableQuery(object):
  """Query for generating results in a table."""

  def __init__(self, fuzzer_name, job_types, stats_columns, group_by,
               date_start, date_end):
    assert group_by

    self.fuzzer_name = fuzzer_name
    self.job_types = job_types
    self.group_by = group_by
    self.date_start = date_start
    self.date_end = date_end
    self.job_run_query = None
    self.testcase_run_query = None

    job_run_fields = []
    testcase_run_fields = []
    fields = parse_stats_column_fields(stats_columns)

    for field in fields:
      # Split up fields by table.
      if not isinstance(field, QueryField):
        continue

      if field.table_alias == JobQuery.ALIAS:
        job_run_fields.append(field)
      elif field.table_alias == TestcaseQuery.ALIAS:
        testcase_run_fields.append(field)

    #  subqueries.

    # For query by time, we can't correlate the time of testcase run with a job
    # run since they are set at different times. So, use only the results from
    # testcase run and don't join them with job run, see build(). Also, the job
    # parameters like: known crashes, new crashes are aggregate numbers from job
    # that are not applicable to show per testcase run (a point on graph).
    if job_run_fields and self.group_by != QueryGroupBy.GROUP_BY_TIME:
      self.job_run_query = JobQuery(fuzzer_name, job_types, job_run_fields,
                                    group_by, date_start, date_end)

    if testcase_run_fields:
      self.testcase_run_query = TestcaseQuery(fuzzer_name, job_types,
                                              testcase_run_fields, group_by,
                                              date_start, date_end)

    assert self.job_run_query or self.testcase_run_query, (
        'Unable to create query.')

  def _join_subqueries(self):
    """Create an inner join for subqueries."""
    result = [
        '(%s) as %s' % (self.job_run_query.build(), self.job_run_query.alias),
        'INNER JOIN',
        '(%s) as %s' % (self.testcase_run_query.build(),
                        self.testcase_run_query.alias), 'ON',
        '{job_alias}.{group_by} = {testcase_alias}.{group_by}'.format(
            job_alias=self.job_run_query.alias,
            testcase_alias=self.testcase_run_query.alias,
            group_by=group_by_to_field_name(self.group_by))
    ]
    return ' '.join(result)

  def _single_subquery(self):
    """Create a single subquery."""
    query = self.job_run_query or self.testcase_run_query
    return '(%s) as %s' % (query.build(), query.alias)

  def build(self):
    """Build the table query."""
    valid_run_query = self.job_run_query or self.testcase_run_query
    result = [
        # We need to do the below to avoid the duplicate column name error.
        'SELECT {0}.{1}, * EXCEPT({1}) FROM'.format(
            valid_run_query.alias,
            group_by_to_field_name(valid_run_query.group_by))
    ]

    if self.job_run_query and self.testcase_run_query:
      result.append(self._join_subqueries())
    else:
      result.append(self._single_subquery())

    return ' '.join(result)


def get_coverage_info(fuzzer, date=None):
  """Returns a CoverageInformation entity for a given fuzzer and date. If date
  is not specified, returns the latest entity available."""
  query = data_types.CoverageInformation.query(
      data_types.CoverageInformation.fuzzer == fuzzer)
  if date:
    # Return info for specific date.
    query = query.filter(data_types.CoverageInformation.date == date)
  else:
    # Return latest.
    query = query.order(-data_types.CoverageInformation.date)

  return query.get()


def get_gcs_stats_path(kind, fuzzer, timestamp):
  """Return gcs path in the format "/bucket/path/to/containing_dir/" for the
  given fuzzer, job, and timestamp or revision."""
  bucket_name = big_query.get_bucket()
  if not bucket_name:
    return None

  datetime_value = datetime.datetime.utcfromtimestamp(timestamp)
  dir_name = data_types.coverage_information_date_to_string(datetime_value)

  path = '/%s/%s/%s/date/%s/' % (bucket_name, fuzzer, kind, dir_name)
  return path


@environment.local_noop
def upload_stats(stats_list, filename=None):
  """Upload the fuzzer run to the bigquery bucket. Assumes that all the stats
  given are for the same fuzzer/job run."""
  if not stats_list:
    logs.log_error('Failed to upload fuzzer stats: empty stats.')
    return

  assert isinstance(stats_list, list)

  bucket_name = big_query.get_bucket()
  if not bucket_name:
    logs.log_error('Failed to upload fuzzer stats: missing bucket name.')
    return

  kind = stats_list[0].kind
  fuzzer = stats_list[0].fuzzer

  # Group all stats for fuzz targets.
  fuzzer_or_engine_name = get_fuzzer_or_engine_name(fuzzer)

  if not filename:
    # Generate a random filename.
    filename = '%016x' % random.randint(0, (1 << 64) - 1) + '.json'

  # Handle runs that bleed into the next day.
  timestamp_start_of_day = lambda s: utils.utc_date_to_timestamp(
      datetime.datetime.utcfromtimestamp(s.timestamp).date())
  stats_list.sort(key=lambda s: s.timestamp)

  for timestamp, stats in itertools.groupby(stats_list, timestamp_start_of_day):
    upload_data = '\n'.join(stat.to_json() for stat in stats)

    day_path = 'gs:/' + get_gcs_stats_path(
        kind, fuzzer_or_engine_name, timestamp=timestamp) + filename

    if not storage.write_data(upload_data.encode('utf-8'), day_path):
      logs.log_error('Failed to upload FuzzerRun.')


def parse_stats_column_fields(column_fields):
  """Parse the stats column fields."""
  # e.g. 'sum(t.field_name) as display_name'.
  aggregate_regex = re.compile(r'^(\w+)\(([a-z])\.([^\)]+)\)(\s*as\s*(\w+))?$')

  # e.g. '_EDGE_COV as blah'.
  builtin_regex = re.compile(r'^(_\w+)(\s*as\s*(\w+))?$')

  fields = []
  parts = [field.strip() for field in column_fields.split(',')]
  for part in parts:
    match = aggregate_regex.match(part)
    if match:
      table_alias = match.group(2)
      field_name = match.group(3)
      aggregate_function = match.group(1)
      select_alias = match.group(5)
      if select_alias:
        select_alias = select_alias.strip('"')

      fields.append(
          QueryField(table_alias, field_name, aggregate_function, select_alias))
      continue

    match = builtin_regex.match(part)
    if match:
      name = match.group(1)
      alias = match.group(3)
      if alias:
        alias = alias.strip('"')
      fields.append(BuiltinFieldSpecifier(name, alias))
      continue

  return fields


def get_fuzzer_or_engine_name(fuzzer_name):
  """Return fuzzing engine name if it exists, or |fuzzer_name|."""
  fuzz_target = data_handler.get_fuzz_target(fuzzer_name)
  if fuzz_target:
    return fuzz_target.engine

  return fuzzer_name


def dataset_name(fuzzer_name):
  """Get the stats dataset name for the given |fuzzer_name|."""
  return fuzzer_name.replace('-', '_') + '_stats'


BUILTIN_FIELD_CONSTRUCTORS = {
    '_EDGE_COV':
        functools.partial(CoverageField, CoverageField.EDGE),
    '_FUNC_COV':
        functools.partial(CoverageField, CoverageField.FUNCTION),
    '_CORPUS_SIZE':
        functools.partial(CorpusSizeField, CorpusSizeField.CORPUS),
    '_CORPUS_BACKUP':
        CorpusBackupField,
    '_QUARANTINE_SIZE':
        functools.partial(CorpusSizeField, CorpusSizeField.QUARANTINE),
    '_COV_REPORT':
        CoverageReportField,
    '_FUZZER_RUN_LOGS':
        FuzzerRunLogsField,
    '_PERFORMANCE_REPORT':
        PerformanceReportField,
}



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""helpers tests."""
# pylint: disable=protected-access

import unittest

from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from libs import auth
from libs import helpers


class TestNotFoundException(Exception):
  """Serve as a testing exception."""


class IsNotEmpty(unittest.TestCase):
  """Test _is_not_empty."""

  def test_empty(self):
    """Test none."""
    self.assertFalse(helpers._is_not_empty(None))
    self.assertFalse(helpers._is_not_empty(''))
    self.assertFalse(helpers._is_not_empty(0))

  def test_not_empty(self):
    """Test not empty."""
    self.assertTrue(helpers._is_not_empty('a'))
    self.assertTrue(helpers._is_not_empty(1))
    self.assertTrue(helpers._is_not_empty(object()))

  def test_tuple_empty(self):
    """Test tuple containing only empty values."""
    self.assertFalse(helpers._is_not_empty((None, None)))
    self.assertFalse(helpers._is_not_empty((None, None, None)))
    self.assertFalse(helpers._is_not_empty(('', '')))
    self.assertFalse(helpers._is_not_empty((None, '', 0, None)))

  def test_tuple_not_empty(self):
    """Test tuples that are not considered as empty."""
    self.assertTrue(helpers._is_not_empty((None, 1)))
    self.assertTrue(helpers._is_not_empty(('a', 0)))


class GetOrExitTest(unittest.TestCase):
  """Test get_or_exit."""

  def test_should_render_json(self):
    """Ensure it determines render_json correctly."""
    self.assertTrue(
        helpers.should_render_json('adsasd;application/json;sdf', ''))
    self.assertTrue(helpers.should_render_json('', 'application/json'))
    self.assertFalse(helpers.should_render_json('text/plain;0.8', 'text/html'))

  def test_get_or_exit_valid(self):
    """Ensure it gets value."""
    fn = lambda: 'test'
    self.assertEqual(helpers.get_or_exit(fn, 'not_found', 'error'), 'test')

  def test_get_or_exit_none(self):
    """Ensure it raises 404 when the value is None."""
    fn = lambda: None
    with self.assertRaises(helpers.EarlyExitException) as catched:
      helpers.get_or_exit(fn, 'not_found', 'error')

    self.assertEqual(catched.exception.status, 404)
    self.assertEqual(str(catched.exception), 'not_found')

  def test_get_or_exit_tuple_none(self):
    """Ensure it raises 404 when the value is a tuple of None."""
    fn = lambda: (None, None)
    with self.assertRaises(helpers.EarlyExitException) as catched:
      helpers.get_or_exit(fn, 'not_found', 'error')

    self.assertEqual(catched.exception.status, 404)
    self.assertEqual(str(catched.exception), 'not_found')

  def test_get_or_exit_not_found_exception(self):
    """Ensure it raises 404 when `fn` throws a recognised exception."""

    def fn():
      raise TestNotFoundException()

    with self.assertRaises(helpers.EarlyExitException) as catched:
      helpers.get_or_exit(
          fn, 'not_found', 'error', not_found_exception=TestNotFoundException)

    self.assertEqual(catched.exception.status, 404)
    self.assertEqual(str(catched.exception), 'not_found')

  def test_get_or_exit_other_exception(self):
    """Ensure it raises 500 when `fn` throws an unknown exception."""

    def fn():
      raise Exception('message')

    with self.assertRaises(helpers.EarlyExitException) as catched:
      helpers.get_or_exit(fn, 'not_found', 'other')

    self.assertEqual(catched.exception.status, 500)

    self.assertEqual(
        str(catched.exception), "other (<class 'Exception'>: message)")


class GetUserEmailTest(unittest.TestCase):
  """Test get_user_email."""

  def setUp(self):
    test_helpers.patch(self, ['libs.auth.get_current_user'])

  def test_get_user_email_success(self):
    """Ensure it gets the email when a user is valid."""
    self.mock.get_current_user.return_value = (auth.User('TeSt@Test.com'))
    self.assertEqual(helpers.get_user_email(), 'TeSt@Test.com')

  def test_get_user_email_failure(self):
    """Ensure it gets empty string when a user is invalid."""
    self.mock.get_current_user.side_effect = Exception()
    self.assertEqual(helpers.get_user_email(), '')


class LogTest(unittest.TestCase):
  """Test log."""

  def setUp(self):
    test_helpers.patch(self, ['logging.info', 'libs.helpers.get_user_email'])
    self.mock.get_user_email.return_value = 'email'

  def test_modify(self):
    """Test log modify."""
    helpers.log('message', helpers.MODIFY_OPERATION)
    self.mock.info.assert_called_once_with('ClusterFuzz: %s (%s): %s.',
                                           helpers.MODIFY_OPERATION, 'email',
                                           'message')

  def test_view(self):
    """Test log view."""
    helpers.log('message', helpers.VIEW_OPERATION)
    self.mock.info.assert_called_once_with(
        'ClusterFuzz: %s (%s): %s.', helpers.VIEW_OPERATION, 'email', 'message')



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""form.py contains static methods around form generation."""
import base64
import datetime
import os

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.datastore import ndb_utils
from libs import helpers


def generate_csrf_token(length=64, valid_seconds=3600, html=False):
  """Generate a CSRF token."""
  now = utils.utcnow()
  valid_token = None

  # Clean up expired tokens to prevent junk from building up in the datastore.
  tokens = data_types.CSRFToken.query(
      data_types.CSRFToken.user_email == helpers.get_user_email())
  tokens_to_delete = []
  for token in tokens:
    if token.expiration_time > now:
      valid_token = token
      continue
    tokens_to_delete.append(token.key)
  ndb_utils.delete_multi(tokens_to_delete)

  # Generate a new token.
  if not valid_token:
    valid_token = data_types.CSRFToken()
    valid_token.value = base64.b64encode(os.urandom(length))
    valid_token.expiration_time = (
        now + datetime.timedelta(seconds=valid_seconds))
    valid_token.user_email = helpers.get_user_email()
    valid_token.put()

  value = valid_token.value
  if html:
    return '<input type="hidden" name="csrf_token" value="%s" />' % value
  return value



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Python 3 entrypoint."""
import importlib
import os
import sys

# Add necessary directories to path.
sys.path.append('third_party')

config_modules_path = os.path.join('config', 'modules')
if os.path.exists(config_modules_path):
  sys.path.append(config_modules_path)

gae_env = os.environ.get('GAE_ENV')
if gae_env:
  import pkg_resources
  importlib.reload(pkg_resources)

  if gae_env != 'dev':
    import firebase_admin
    firebase_admin.initialize_app()

try:
  # Run any module initialization code.
  import module_init
  module_init.appengine()
except ImportError:
  pass

import server

app = server.app



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Protobuf helpers."""


def get_protobuf_field(result, buf, field_name):
  """Add protobuf field to dict, if the field exists."""
  if buf.HasField(field_name):
    result[field_name] = getattr(buf, field_name)


def encode_utf8_if_unicode(data):
  """Encode string as utf-8 if it's unicode."""
  if isinstance(data, str):
    return data.encode('utf-8')

  return data



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Used to download Android symbols."""

import os

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.google_cloud_utils import storage
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.platforms.android import adb
from clusterfuzz._internal.platforms.android import fetch_artifact
from clusterfuzz._internal.platforms.android import kernel_utils
from clusterfuzz._internal.platforms.android import settings
from clusterfuzz._internal.system import archive
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import shell


def get_repo_prop_archive_filename(build_id, target):
  return f'{target}-{build_id}-repo.prop'


def should_download_symbols():
  """Return True if we should continue to download symbols."""
  # For local testing or when running the reproduce tool locally, we do not
  # have access to the cloud storage bucket with the symbols. In this case,
  # just bail out.
  # We have archived symbols for google builds only.
  return (not environment.get_value('LOCAL_DEVELOPMENT') and
          not environment.get_value('REPRODUCE_TOOL') and
          settings.is_google_device())


def download_artifact_if_needed(
    build_id, artifact_directory, artifact_archive_path,
    targets_with_type_and_san, artifact_file_name, output_filename_override,
    build_params, build_params_check_path):
  """Downloads artifact to actifacts_archive_path if needed"""
  # Check if we already have the symbols in cache.
  cached_build_params = utils.read_data_from_file(
      build_params_check_path, eval_data=True)
  if cached_build_params and cached_build_params == build_params:
    # No work to do, same system symbols already in cache.
    return

  # Delete existing symbols directory first.
  shell.remove_directory(artifact_directory, recreate=True)

  # Fetch symbol file from cloud storage cache (if available).
  found_in_cache = storage.get_file_from_cache_if_exists(
      artifact_archive_path, update_modification_time_on_access=False)
  if not found_in_cache:
    for target_with_type_and_san in targets_with_type_and_san:
      # Fetch the artifact now.
      fetch_artifact.get(build_id, target_with_type_and_san, artifact_file_name,
                         artifact_directory, output_filename_override)
      if os.path.exists(artifact_archive_path):
        break


def download_repo_prop_if_needed(symbols_directory, build_id, cache_target,
                                 targets_with_type_and_san, cache_type):
  """Downloads the repo.prop for a branch"""
  artifact_file_name = 'repo.prop'
  symbols_archive_filename = get_repo_prop_archive_filename(
      build_id, cache_target)
  output_filename_override = symbols_archive_filename
  # We create our own build_params for cache
  build_params = {
      'build_id': build_id,
      'target': cache_target,
      'type': cache_type
  }

  build_params_check_path = os.path.join(symbols_directory,
                                         '.cached_build_params')

  symbols_archive_path = os.path.join(symbols_directory,
                                      symbols_archive_filename)
  download_artifact_if_needed(build_id, symbols_directory, symbols_archive_path,
                              targets_with_type_and_san, artifact_file_name,
                              output_filename_override, build_params,
                              build_params_check_path)
  if not os.path.exists(symbols_archive_path):
    logs.log_error('Unable to locate repo.prop %s.' % symbols_archive_path)
    return

  # Store the artifact for later use or for use by other bots.
  storage.store_file_in_cache(symbols_archive_path)
  utils.write_data_to_file(build_params, build_params_check_path)


def download_kernel_repo_prop_if_needed(symbols_directory):
  """Downloads the repo.prop for the kernel of a device"""
  if not should_download_symbols():
    return

  # For Android kernel we want to get the repro.prop
  # Note: kasan and non-kasan kernel should have the same repo.prop for a given
  # build_id.
  _, build_id = kernel_utils.get_kernel_hash_and_build_id()
  target = kernel_utils.get_kernel_name()
  if not build_id or not target:
    logs.log_error('Could not get kernel parameters, exiting.')
    return

  tool_suffix = environment.get_value('SANITIZER_TOOL_NAME')
  # Some kernels are just 'kernel', some are kernel_target
  if tool_suffix:
    targets_with_type_and_san = [
        f'kernel_{tool_suffix}', f'kernel_{tool_suffix}_{target}'
    ]
  else:
    targets_with_type_and_san = ['kernel', f'kernel_{target}']

  download_repo_prop_if_needed(symbols_directory, build_id, target,
                               targets_with_type_and_san, 'kernel')


def download_system_symbols_if_needed(symbols_directory):
  """Download system libraries from |SYMBOLS_URL| and cache locally."""
  if not should_download_symbols():
    return

  # Get the build fingerprint parameters.
  build_params = settings.get_build_parameters()
  if not build_params:
    logs.log_error('Unable to determine build parameters.')
    return

  build_params_check_path = os.path.join(symbols_directory,
                                         '.cached_build_params')
  build_id = build_params.get('build_id')
  target = build_params.get('target')
  build_type = build_params.get('type')
  if not build_id or not target or not build_type:
    logs.log_error('Null build parameters found, exiting.')
    return

  symbols_archive_filename = f'{target}-symbols-{build_id}.zip'
  artifact_file_name = symbols_archive_filename
  output_filename_override = None

  # Include type and sanitizer information in the target.
  tool_suffix = environment.get_value('SANITIZER_TOOL_NAME')
  target_with_type_and_san = f'{target}-{build_type}'
  if tool_suffix and not tool_suffix in target_with_type_and_san:
    target_with_type_and_san += f'_{tool_suffix}'

  targets_with_type_and_san = [target_with_type_and_san]

  symbols_archive_path = os.path.join(symbols_directory,
                                      symbols_archive_filename)
  download_artifact_if_needed(build_id, symbols_directory, symbols_archive_path,
                              targets_with_type_and_san, artifact_file_name,
                              output_filename_override, build_params,
                              build_params_check_path)
  if not os.path.exists(symbols_archive_path):
    logs.log_error(
        'Unable to locate symbols archive %s.' % symbols_archive_path)
    return

  # Store the artifact for later use or for use by other bots.
  storage.store_file_in_cache(symbols_archive_path)

  archive.unpack(symbols_archive_path, symbols_directory, trusted=True)
  shell.remove_file(symbols_archive_path)

  utils.write_data_to_file(build_params, build_params_check_path)


def _get_binary_from_build_or_device(binary_path):
  """Look for binary on build server or on device."""
  # Initialize some helper variables.
  symbols_directory = environment.get_value('SYMBOLS_DIR')
  binary_filename = os.path.basename(binary_path)

  # We didn't find the library locally in the build directory.
  # Try finding the library in the local system library cache.
  download_system_symbols_if_needed(symbols_directory)
  local_binary_path = utils.find_binary_path(symbols_directory, binary_path)
  if local_binary_path:
    return local_binary_path

  # Try pulling in the binary directly from the device into the
  # system library cache directory.
  local_binary_path = os.path.join(symbols_directory, binary_filename)
  adb.run_command('pull %s %s' % (binary_path, local_binary_path))
  if os.path.exists(local_binary_path):
    return local_binary_path

  return None


def filter_binary_path(binary_path):
  """Filter binary path to provide local copy."""
  # LKL fuzzer name is not full path.
  if environment.is_android():
    # Skip symbolization when running it on bad entries like [stack:XYZ].
    if not binary_path.startswith('/') or '(deleted)' in binary_path:
      return ''

  # Initialize some helper variables.
  build_directory = environment.get_value('BUILD_DIR')

  # Try to find the library in the build directory first.
  local_binary_path = utils.find_binary_path(build_directory, binary_path)
  if local_binary_path:
    return local_binary_path

  # We should only download from the build server if we are Android.
  if environment.is_android():
    local_binary_path = _get_binary_from_build_or_device(binary_path)
    if local_binary_path:
      return local_binary_path

  # Unable to find library.
  logs.log_error('Unable to find library %s for symbolization.' % binary_path)
  return ''



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""manage_vms tests."""

import copy
import functools
import unittest

from google.cloud import ndb
import mock
import six

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.google_cloud_utils import compute_engine_projects
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from handlers.cron import manage_vms
from handlers.cron.helpers import bot_manager

INSTANCE_GROUPS = {
    'oss-fuzz-linux-zone2-pre-proj2': {
        'targetSize': 1,
    },
    'oss-fuzz-linux-zone2-pre-proj3': {
        'targetSize': 499,
    },
    'oss-fuzz-linux-zone2-pre-proj4': {
        'targetSize': 99,
    },
    'oss-fuzz-linux-zone2-pre-proj5': {
        'targetSize': 99,
    }
}

INSTANCE_TEMPLATES = {
    'oss-fuzz-linux-zone2-pre-proj2': {
        'description': '{"version": 1}',
        'properties': {
            'metadata': {
                'items': [],
            },
            'disks': [{
                'initializeParams': {
                    'diskSizeGb': '30',
                },
            }],
            'serviceAccounts': [{
                'email':
                    'email',
                'scopes': [
                    'https://www.googleapis.com/auth/'
                    'devstorage.full_control',
                    'https://www.googleapis.com/auth/logging.write',
                    'https://www.googleapis.com/auth/userinfo.email',
                    'https://www.googleapis.com/auth/appengine.apis',
                    'https://www.googleapis.com/auth/prodxmon',
                    'https://www.googleapis.com/auth/bigquery',
                ]
            }],
        }
    },
    'oss-fuzz-linux-zone2-pre-proj3': {
        'description': '{"version": 1}',
        'properties': {
            'metadata': {
                'items': [],
            },
            'disks': [{
                'initializeParams': {
                    'diskSizeGb': '30',
                },
            }],
            'serviceAccounts': [{
                'email':
                    'email',
                'scopes': [
                    'https://www.googleapis.com/auth/'
                    'devstorage.full_control',
                    'https://www.googleapis.com/auth/logging.write',
                    'https://www.googleapis.com/auth/userinfo.email',
                    'https://www.googleapis.com/auth/appengine.apis',
                    'https://www.googleapis.com/auth/prodxmon',
                    'https://www.googleapis.com/auth/bigquery',
                ]
            }],
        }
    },
    'oss-fuzz-linux-zone2-pre-proj4': {
        'description': '{"version": 0}',
        'properties': {
            'metadata': {
                'items': [],
            },
            'disks': [{
                'initializeParams': {
                    'diskSizeGb': '30',
                },
            }],
            'serviceAccounts': [{
                'email':
                    'email',
                'scopes': [
                    'https://www.googleapis.com/auth/'
                    'devstorage.full_control',
                    'https://www.googleapis.com/auth/logging.write',
                    'https://www.googleapis.com/auth/userinfo.email',
                    'https://www.googleapis.com/auth/appengine.apis',
                    'https://www.googleapis.com/auth/prodxmon',
                    'https://www.googleapis.com/auth/bigquery',
                ]
            }],
        }
    },
    'oss-fuzz-linux-zone2-pre-proj5': {
        'description': '{"version": 1}',
        'properties': {
            'metadata': {
                'items': [],
            },
            'disks': [{
                'initializeParams': {
                    'diskSizeGb': '30',
                },
            }],
            'serviceAccounts': [{
                'email':
                    'email',
                'scopes': [
                    'https://www.googleapis.com/auth/'
                    'devstorage.full_control',
                    'https://www.googleapis.com/auth/logging.write',
                    'https://www.googleapis.com/auth/userinfo.email',
                    'https://www.googleapis.com/auth/appengine.apis',
                    'https://www.googleapis.com/auth/prodxmon',
                    'https://www.googleapis.com/auth/bigquery',
                ]
            }],
        }
    }
}

INSTANCES = {
    'oss-fuzz-linux-zone3-host': [{
        'instance': 'https://blah/oss-fuzz-linux-zone3-host-abcd',
    }, {
        'instance': 'https://blah/oss-fuzz-linux-zone3-host-efgh',
    }],
    'oss-fuzz-linux-zone3-worker-proj1': [{
        'instance': 'https://blah/oss-fuzz-linux-zone3-worker-proj1-%04d' % i
    } for i in range(1, 2)],
    'oss-fuzz-linux-zone3-worker-proj2': [{
        'instance': 'https://blah/oss-fuzz-linux-zone3-worker-proj2-%04d' % i
    } for i in range(1, 5)],
    'oss-fuzz-linux-zone3-worker-proj3': [{
        'instance': 'https://blah/oss-fuzz-linux-zone3-worker-proj3-%04d' % i
    } for i in range(1, 10)],
    'oss-fuzz-linux-zone3-worker-proj4': [{
        'instance': 'https://blah/oss-fuzz-linux-zone3-worker-proj4-%04d' % i
    } for i in range(1, 2)],
    'oss-fuzz-linux-zone3-worker-proj5': [{
        'instance': 'https://blah/oss-fuzz-linux-zone3-worker-proj5-%04d' % i
    } for i in range(1, 2)],
    'oss-fuzz-linux-zone3-host-high-end': [{
        'instance': 'https://blah/oss-fuzz-linux-zone3-host-high-end-1'
    }],
    'oss-fuzz-linux-zone3-worker-high-end-proj6': [{
        'instance': ('https://blah/'
                     'oss-fuzz-linux-zone3-worker-high-end-proj6-%04d' % i)
    } for i in range(1, 3)],
}

OSS_FUZZ_CLUSTERS = compute_engine_projects.Project(
    project_id='clusterfuzz-external',
    clusters=[
        compute_engine_projects.Cluster(
            name='oss-fuzz-linux-zone2-pre',
            gce_zone='us-east2-a',
            instance_count=997,
            instance_template='external-pre-zone2',
            distribute=True,
            worker=False,
            high_end=False),
        compute_engine_projects.Cluster(
            name='oss-fuzz-linux-zone3-host',
            gce_zone='us-central1-d',
            instance_count=2,
            instance_template='host-zone3',
            distribute=False,
            worker=False,
            high_end=False),
        compute_engine_projects.Cluster(
            name='oss-fuzz-linux-zone3-worker',
            gce_zone='us-central1-d',
            instance_count=16,
            instance_template='worker-zone3',
            distribute=True,
            worker=True,
            high_end=False),
        compute_engine_projects.Cluster(
            name='oss-fuzz-linux-zone3-host-high-end',
            gce_zone='us-central1-d',
            instance_count=1,
            instance_template='host-high-end-zone3',
            distribute=False,
            worker=False,
            high_end=True),
        compute_engine_projects.Cluster(
            name='oss-fuzz-linux-zone3-worker-high-end',
            gce_zone='us-central1-d',
            instance_count=2,
            instance_template='worker-zone3',
            distribute=True,
            worker=True,
            high_end=True),
    ],
    instance_templates=[
        {
            'name': 'external-pre-zone2',
            'description': '{"version": 1}',
            'properties': {
                'metadata': {
                    'items': [],
                },
                'disks': [{
                    'initializeParams': {
                        'diskSizeGb': 30,
                    },
                }],
                'serviceAccounts': [{
                    'email':
                        'email',
                    'scopes': [
                        'https://www.googleapis.com/auth/'
                        'devstorage.full_control',
                        'https://www.googleapis.com/auth/logging.write',
                        'https://www.googleapis.com/auth/userinfo.email',
                        'https://www.googleapis.com/auth/appengine.apis',
                        'https://www.googleapis.com/auth/prodxmon',
                        'https://www.googleapis.com/auth/bigquery',
                    ]
                }],
            }
        },
        {
            'name': 'host-zone3',
            'description': '{"version": 1}',
            'properties': {
                'metadata': {
                    'items': [],
                },
                'disks': [{
                    'initializeParams': {
                        'diskSizeGb': 30,
                    },
                }],
                'serviceAccounts': [{
                    'email':
                        'email',
                    'scopes': [
                        'https://www.googleapis.com/auth/'
                        'devstorage.full_control',
                        'https://www.googleapis.com/auth/logging.write',
                        'https://www.googleapis.com/auth/userinfo.email',
                        'https://www.googleapis.com/auth/appengine.apis',
                        'https://www.googleapis.com/auth/prodxmon',
                        'https://www.googleapis.com/auth/bigquery',
                    ]
                }],
            }
        },
        {
            'name': 'worker-zone3',
            'description': '{"version": 1}',
            'properties': {
                'metadata': {
                    'items': [],
                },
                'disks': [{
                    'initializeParams': {
                        'diskSizeGb': 30,
                    },
                }],
                'serviceAccounts': [{
                    'email':
                        'email',
                    'scopes': [
                        'https://www.googleapis.com/auth/'
                        'devstorage.full_control',
                        'https://www.googleapis.com/auth/logging.write',
                        'https://www.googleapis.com/auth/userinfo.email',
                        'https://www.googleapis.com/auth/prodxmon',
                    ]
                }],
            }
        },
        {
            'name': 'host-high-end-zone3',
            'description': '{"version": 1}',
            'properties': {
                'metadata': {
                    'items': [],
                },
                'disks': [{
                    'initializeParams': {
                        'diskSizeGb': 100,
                    },
                }],
                'serviceAccounts': [{
                    'email':
                        'email',
                    'scopes': [
                        'https://www.googleapis.com/auth/'
                        'devstorage.full_control',
                        'https://www.googleapis.com/auth/logging.write',
                        'https://www.googleapis.com/auth/userinfo.email',
                        'https://www.googleapis.com/auth/prodxmon',
                    ]
                }],
            }
        },
    ],
    host_worker_assignments=[
        compute_engine_projects.HostWorkerAssignment(
            host='oss-fuzz-linux-zone3-host',
            worker='oss-fuzz-linux-zone3-worker',
            workers_per_host=8),
        compute_engine_projects.HostWorkerAssignment(
            host='oss-fuzz-linux-zone3-host-high-end',
            worker='oss-fuzz-linux-zone3-worker-high-end',
            workers_per_host=2),
    ])


def mock_resource(spec):
  """Mock resource."""
  resource = mock.Mock(spec=spec)
  resource.created = False
  resource.body = None

  def create(*args, **kwargs):  # pylint: disable=unused-argument
    if resource.created:
      raise bot_manager.AlreadyExistsError

    resource.created = True

  def get():
    if resource.created:
      return resource.body

    raise bot_manager.NotFoundError

  def exists():
    return resource.created

  def delete():
    if not resource.created:
      raise bot_manager.NotFoundError

    resource.created = False

  resource.create.side_effect = create
  resource.get.side_effect = get
  resource.exists.side_effect = exists
  resource.delete.side_effect = delete

  return resource


class MockBotManager(object):
  """Mock BotManager."""

  def __init__(self, project_id, zone, instance_groups, instance_templates):
    self.project_id = project_id
    self.zone = zone
    self.instance_groups = instance_groups
    self.instance_templates = instance_templates

  def _get_resource(self, name, cache, values, spec):
    """Get resource."""
    if name in cache:
      return cache[name]

    resource = mock_resource(spec=spec)
    if name in values:
      resource.created = True
      resource.body = values[name]

    cache[name] = resource
    return resource

  def instance_group(self, name):
    """Get an InstanceGroup resource with the given name."""
    resource = self._get_resource(name, self.instance_groups, INSTANCE_GROUPS,
                                  bot_manager.InstanceGroup)

    if name in INSTANCES:
      resource.list_managed_instances.return_value = INSTANCES[name]

    return resource

  def instance_template(self, name):
    """Get an InstanceTemplate resource with the given name."""
    return self._get_resource(name, self.instance_templates, INSTANCE_TEMPLATES,
                              bot_manager.InstanceTemplate)


def expected_instance_template(gce_project_name,
                               name,
                               project_name,
                               disk_size_gb=None,
                               service_account=None,
                               tls_cert=False):
  """Get the expected instance template for a project."""
  gce_project = compute_engine_projects.load_project(gce_project_name)
  expected = copy.deepcopy(gce_project.get_instance_template(name))
  expected['properties']['metadata']['items'].append({
      'key': 'task-tag',
      'value': project_name,
  })

  if disk_size_gb:
    disk = expected['properties']['disks'][0]
    disk['initializeParams']['diskSizeGb'] = disk_size_gb

  if service_account:
    expected['properties']['serviceAccounts'][0]['email'] = service_account

  if tls_cert:
    expected['properties']['metadata']['items'].extend([{
        'key': 'tls-cert',
        'value': project_name + '_cert',
    }, {
        'key': 'tls-key',
        'value': project_name + '_key',
    }])

  return expected


def expected_host_instance_template(gce_project_name, name):
  """Get the expected instance template for a project."""
  gce_project = compute_engine_projects.load_project(gce_project_name)
  return copy.deepcopy(gce_project.get_instance_template(name))


@test_utils.with_cloud_emulators('datastore')
class CronTest(unittest.TestCase):
  """Test manage_vms cron."""

  def setUp(self):
    test_helpers.patch_environ(self)
    test_helpers.patch(self, [
        'clusterfuzz._internal.base.utils.is_oss_fuzz',
        'handlers.cron.helpers.bot_manager.BotManager',
        'clusterfuzz._internal.system.environment.is_running_on_app_engine',
        'clusterfuzz._internal.google_cloud_utils.compute_engine_projects.load_project',
    ])

    self.mock.is_oss_fuzz.return_value = True
    self.mock.is_running_on_app_engine.return_value = True
    self.mock.load_project.return_value = OSS_FUZZ_CLUSTERS

    data_types.OssFuzzProject(
        id='proj1',
        name='proj1',
        cpu_weight=1.0,
        service_account='proj1@serviceaccount.com').put()

    data_types.OssFuzzProject(
        id='proj2',
        name='proj2',
        cpu_weight=2.0,
        service_account='proj2@serviceaccount.com').put()

    data_types.OssFuzzProject(
        id='proj3',
        name='proj3',
        cpu_weight=5.0,
        service_account='proj3@serviceaccount.com').put()

    data_types.OssFuzzProject(
        id='proj4',
        name='proj4',
        cpu_weight=1.0,
        service_account='proj4@serviceaccount.com').put()

    data_types.OssFuzzProject(
        id='proj5',
        name='proj5',
        cpu_weight=1.0,
        service_account='proj5@serviceaccount.com',
        disk_size_gb=10).put()

    data_types.OssFuzzProject(
        id='proj6',
        name='proj6',
        cpu_weight=1.0,
        service_account='proj6@serviceaccount.com',
        high_end=True).put()

    for j in range(1, 7):
      project_name = 'proj%d' % j
      data_types.WorkerTlsCert(
          id=project_name,
          project_name=project_name,
          cert_contents=project_name.encode() + b'_cert',
          key_contents=project_name.encode() + b'_key').put()

    data_types.OssFuzzProjectInfo(id='old_proj', name='old_proj').put()

    data_types.OssFuzzProjectInfo(
        id='proj2',
        name='proj2',
        clusters=[
            data_types.OssFuzzProjectInfo.ClusterInfo(
                cluster='oss-fuzz-linux-zone2-pre',
                gce_zone='us-east2-a',
                cpu_count=1,
            ),
            data_types.OssFuzzProjectInfo.ClusterInfo(
                cluster='old-cluster',
                gce_zone='us-east2-a',
                cpu_count=1,
            ),
        ]).put()

    data_types.OssFuzzProjectInfo(
        id='proj3',
        name='proj3',
        clusters=[
            data_types.OssFuzzProjectInfo.ClusterInfo(
                cluster='oss-fuzz-linux-zone2-pre',
                gce_zone='us-east2-a',
                cpu_count=499,
            )
        ]).put()

    data_types.OssFuzzProjectInfo(
        id='proj4',
        name='proj4',
        clusters=[
            data_types.OssFuzzProjectInfo.ClusterInfo(
                cluster='oss-fuzz-linux-zone2-pre',
                gce_zone='us-east2-a',
                cpu_count=99,
            )
        ]).put()

    data_types.OssFuzzProjectInfo(
        id='proj5',
        name='proj5',
        clusters=[
            data_types.OssFuzzProjectInfo.ClusterInfo(
                cluster='oss-fuzz-linux-zone2-pre',
                gce_zone='us-east2-a',
                cpu_count=99,
            )
        ]).put()

    data_types.OssFuzzProjectInfo(
        id='old_proj',
        name='old_proj',
        clusters=[
            data_types.OssFuzzProjectInfo.ClusterInfo(
                cluster='oss-fuzz-linux-zone2-pre',
                gce_zone='us-east2-a',
                cpu_count=5,
            )
        ]).put()

    data_types.HostWorkerAssignment(
        id='old-host-0',
        host_name='old-host',
        worker_name='worker',
        instance_num=0).put()

    instance_groups = {}
    instance_templates = {}
    self.mock.BotManager.side_effect = functools.partial(
        MockBotManager,
        instance_groups=instance_groups,
        instance_templates=instance_templates)

  def test_update_cpus(self):
    """Tests CPU distribution cron."""
    self.maxDiff = None  # pylint: disable=invalid-name
    manager = manage_vms.OssFuzzClustersManager('clusterfuzz-external')
    manager.update_clusters()

    proj1 = ndb.Key(data_types.OssFuzzProjectInfo, 'proj1').get()
    self.assertIsNotNone(proj1)
    self.assertDictEqual({
        'name':
            'proj1',
        'clusters': [{
            'cluster': 'oss-fuzz-linux-zone2-pre',
            'cpu_count': 100,
            'gce_zone': 'us-east2-a',
        }, {
            'cluster': 'oss-fuzz-linux-zone3-worker',
            'cpu_count': 1,
            'gce_zone': 'us-central1-d',
        }],
    }, proj1.to_dict())

    proj2 = ndb.Key(data_types.OssFuzzProjectInfo, 'proj2').get()
    self.assertIsNotNone(proj2)
    self.assertDictEqual({
        'name':
            'proj2',
        'clusters': [{
            'cluster': 'oss-fuzz-linux-zone2-pre',
            'cpu_count': 200,
            'gce_zone': 'us-east2-a',
        }, {
            'cluster': 'oss-fuzz-linux-zone3-worker',
            'cpu_count': 4,
            'gce_zone': 'us-central1-d',
        }],
    }, proj2.to_dict())

    proj3 = ndb.Key(data_types.OssFuzzProjectInfo, 'proj3').get()
    self.assertIsNotNone(proj3)
    self.assertDictEqual({
        'name':
            'proj3',
        'clusters': [{
            'cluster': 'oss-fuzz-linux-zone2-pre',
            'cpu_count': 499,
            'gce_zone': 'us-east2-a',
        }, {
            'cluster': 'oss-fuzz-linux-zone3-worker',
            'cpu_count': 9,
            'gce_zone': 'us-central1-d',
        }],
    }, proj3.to_dict())

    proj4 = ndb.Key(data_types.OssFuzzProjectInfo, 'proj4').get()
    self.assertIsNotNone(proj4)
    self.assertDictEqual({
        'name':
            'proj4',
        'clusters': [{
            'cluster': 'oss-fuzz-linux-zone2-pre',
            'cpu_count': 99,
            'gce_zone': 'us-east2-a',
        }, {
            'cluster': 'oss-fuzz-linux-zone3-worker',
            'cpu_count': 1,
            'gce_zone': 'us-central1-d',
        }],
    }, proj4.to_dict())

    proj5 = ndb.Key(data_types.OssFuzzProjectInfo, 'proj5').get()
    self.assertIsNotNone(proj5)
    self.assertDictEqual({
        'name':
            'proj5',
        'clusters': [{
            'cluster': 'oss-fuzz-linux-zone2-pre',
            'cpu_count': 99,
            'gce_zone': 'us-east2-a',
        }, {
            'cluster': 'oss-fuzz-linux-zone3-worker',
            'cpu_count': 1,
            'gce_zone': 'us-central1-d',
        }],
    }, proj5.to_dict())

    proj6 = ndb.Key(data_types.OssFuzzProjectInfo, 'proj6').get()
    self.assertIsNotNone(proj6)
    self.assertDictEqual({
        'name':
            'proj6',
        'clusters': [{
            'cluster': 'oss-fuzz-linux-zone3-worker-high-end',
            'cpu_count': 2,
            'gce_zone': 'us-central1-d',
        }],
    }, proj6.to_dict())

    old_proj = ndb.Key(data_types.OssFuzzProjectInfo, 'old_proj').get()
    self.assertIsNone(old_proj)

    mock_bot_manager = self.mock.BotManager('clusterfuzz-external',
                                            'us-east2-a')

    # proj1: new project.
    mock_bot_manager.instance_template(
        'oss-fuzz-linux-zone2-pre-proj1').create.assert_called_with(
            expected_instance_template('clusterfuzz-external',
                                       'external-pre-zone2', 'proj1'))
    mock_bot_manager.instance_group(
        'oss-fuzz-linux-zone2-pre-proj1').create.assert_called_with(
            'oss-fuzz-linux-zone2-pre-proj1',
            'oss-fuzz-linux-zone2-pre-proj1',
            size=100,
            wait_for_instances=False)
    mock_bot_manager.instance_group(
        'oss-fuzz-linux-zone2-pre-proj1').resize.assert_not_called()

    # proj2: already exists. needs a resize. old cluster should be deleted.
    mock_bot_manager.instance_template(
        'oss-fuzz-linux-zone2-pre-proj2').create.assert_not_called()
    mock_bot_manager.instance_template(
        'oss-fuzz-linux-zone2-pre-proj2').delete.assert_not_called()
    mock_bot_manager.instance_group(
        'oss-fuzz-linux-zone2-pre-proj2').create.assert_not_called()
    mock_bot_manager.instance_group(
        'oss-fuzz-linux-zone2-pre-proj2').delete.assert_not_called()
    mock_bot_manager.instance_group(
        'oss-fuzz-linux-zone2-pre-proj2').resize.assert_called_with(
            200, wait_for_instances=False)
    mock_bot_manager.instance_template(
        'old-cluster-proj2').delete.assert_called()
    mock_bot_manager.instance_group('old-cluster-proj2').delete.assert_called()

    # proj3: already exists. no changes needed.
    mock_bot_manager.instance_template(
        'oss-fuzz-linux-zone2-pre-proj3').delete.assert_not_called()
    mock_bot_manager.instance_template(
        'oss-fuzz-linux-zone2-pre-proj3').create.assert_not_called()
    mock_bot_manager.instance_group(
        'oss-fuzz-linux-zone2-pre-proj3').create.assert_not_called()
    mock_bot_manager.instance_group(
        'oss-fuzz-linux-zone2-pre-proj3').resize.assert_not_called()
    mock_bot_manager.instance_group(
        'oss-fuzz-linux-zone2-pre-proj3').delete.assert_not_called()

    # proj4: needs a template update (version change).
    mock_bot_manager.instance_template(
        'oss-fuzz-linux-zone2-pre-proj4').delete.assert_called()
    mock_bot_manager.instance_template(
        'oss-fuzz-linux-zone2-pre-proj4').create.assert_called_with(
            expected_instance_template('clusterfuzz-external',
                                       'external-pre-zone2', 'proj4'))
    mock_bot_manager.instance_group(
        'oss-fuzz-linux-zone2-pre-proj4').delete.assert_called()
    mock_bot_manager.instance_group(
        'oss-fuzz-linux-zone2-pre-proj4').create.assert_called_with(
            'oss-fuzz-linux-zone2-pre-proj4',
            'oss-fuzz-linux-zone2-pre-proj4',
            size=99,
            wait_for_instances=False)
    mock_bot_manager.instance_group(
        'oss-fuzz-linux-zone2-pre-proj4').resize.assert_not_called()

    # proj5: needs a template update (disk size change).
    mock_bot_manager.instance_template(
        'oss-fuzz-linux-zone2-pre-proj5').delete.assert_called()
    mock_bot_manager.instance_template(
        'oss-fuzz-linux-zone2-pre-proj5').create.assert_called_with(
            expected_instance_template(
                'clusterfuzz-external',
                'external-pre-zone2',
                'proj5',
                disk_size_gb=10))
    mock_bot_manager.instance_group(
        'oss-fuzz-linux-zone2-pre-proj5').delete.assert_called()
    mock_bot_manager.instance_group(
        'oss-fuzz-linux-zone2-pre-proj5').create.assert_called_with(
            'oss-fuzz-linux-zone2-pre-proj5',
            'oss-fuzz-linux-zone2-pre-proj5',
            size=99,
            wait_for_instances=False)
    mock_bot_manager.instance_group(
        'oss-fuzz-linux-zone2-pre-proj5').resize.assert_not_called()

    # proj6: high end project.
    for j in range(1, 6):
      mock_bot_manager.instance_group(
          'oss-fuzz-linux-zone3-worker-high-end-proj' +
          str(j)).create.assert_not_called()

    mock_bot_manager.instance_group(
        'oss-fuzz-linux-zone3-worker-high-end-proj6').create.assert_called()

    # old_proj: deleted.
    mock_bot_manager.instance_group(
        'oss-fuzz-linux-zone2-pre-old-proj').create.assert_not_called()
    mock_bot_manager.instance_template(
        'oss-fuzz-linux-zone2-pre-old-proj').delete.assert_called()
    mock_bot_manager.instance_group(
        'oss-fuzz-linux-zone2-pre-old-proj').delete.assert_called()

    # host instances: created.
    mock_bot_manager.instance_template(
        'oss-fuzz-linux-zone3-host').create.assert_called_with(
            expected_host_instance_template('clusterfuzz-external',
                                            'host-zone3'))
    mock_bot_manager.instance_group(
        'oss-fuzz-linux-zone3-host').create.assert_called_with(
            'oss-fuzz-linux-zone3-host',
            'oss-fuzz-linux-zone3-host',
            size=2,
            wait_for_instances=False)

    mock_bot_manager.instance_group(
        'oss-fuzz-linux-zone3-host-high-end').create.assert_called_with(
            'oss-fuzz-linux-zone3-host-high-end',
            'oss-fuzz-linux-zone3-host-high-end',
            size=1,
            wait_for_instances=False)

    # Worker instances: created.
    mock_bot_manager.instance_template(
        'oss-fuzz-linux-zone3-worker-proj1').create.assert_called_with(
            expected_instance_template(
                'clusterfuzz-external',
                'worker-zone3',
                'proj1',
                service_account='proj1@serviceaccount.com',
                tls_cert=True))
    mock_bot_manager.instance_group(
        'oss-fuzz-linux-zone3-worker-proj1').create.assert_called_with(
            'oss-fuzz-linux-zone3-worker-proj1',
            'oss-fuzz-linux-zone3-worker-proj1',
            size=1,
            wait_for_instances=False)

    mock_bot_manager.instance_template(
        'oss-fuzz-linux-zone3-worker-proj2').create.assert_called_with(
            expected_instance_template(
                'clusterfuzz-external',
                'worker-zone3',
                'proj2',
                service_account='proj2@serviceaccount.com',
                tls_cert=True))
    mock_bot_manager.instance_group(
        'oss-fuzz-linux-zone3-worker-proj2').create.assert_called_with(
            'oss-fuzz-linux-zone3-worker-proj2',
            'oss-fuzz-linux-zone3-worker-proj2',
            size=4,
            wait_for_instances=False)

    six.assertCountEqual(self, [{
        'instance_num': 0,
        'worker_name': u'oss-fuzz-linux-zone3-worker-proj1-0001',
        'project_name': u'proj1',
        'host_name': u'oss-fuzz-linux-zone3-host-abcd'
    }, {
        'instance_num': 1,
        'worker_name': u'oss-fuzz-linux-zone3-worker-proj2-0001',
        'project_name': u'proj2',
        'host_name': u'oss-fuzz-linux-zone3-host-abcd'
    }, {
        'instance_num': 2,
        'worker_name': u'oss-fuzz-linux-zone3-worker-proj2-0002',
        'project_name': u'proj2',
        'host_name': u'oss-fuzz-linux-zone3-host-abcd'
    }, {
        'instance_num': 3,
        'worker_name': u'oss-fuzz-linux-zone3-worker-proj2-0003',
        'project_name': u'proj2',
        'host_name': u'oss-fuzz-linux-zone3-host-abcd'
    }, {
        'instance_num': 4,
        'worker_name': u'oss-fuzz-linux-zone3-worker-proj2-0004',
        'project_name': u'proj2',
        'host_name': u'oss-fuzz-linux-zone3-host-abcd'
    }, {
        'instance_num': 5,
        'worker_name': u'oss-fuzz-linux-zone3-worker-proj3-0001',
        'project_name': u'proj3',
        'host_name': u'oss-fuzz-linux-zone3-host-abcd'
    }, {
        'instance_num': 6,
        'worker_name': u'oss-fuzz-linux-zone3-worker-proj3-0002',
        'project_name': u'proj3',
        'host_name': u'oss-fuzz-linux-zone3-host-abcd'
    }, {
        'instance_num': 7,
        'worker_name': u'oss-fuzz-linux-zone3-worker-proj3-0003',
        'project_name': u'proj3',
        'host_name': u'oss-fuzz-linux-zone3-host-abcd'
    }, {
        'instance_num': 0,
        'worker_name': u'oss-fuzz-linux-zone3-worker-proj3-0004',
        'project_name': u'proj3',
        'host_name': u'oss-fuzz-linux-zone3-host-efgh'
    }, {
        'instance_num': 1,
        'worker_name': u'oss-fuzz-linux-zone3-worker-proj3-0005',
        'project_name': u'proj3',
        'host_name': u'oss-fuzz-linux-zone3-host-efgh'
    }, {
        'instance_num': 2,
        'worker_name': u'oss-fuzz-linux-zone3-worker-proj3-0006',
        'project_name': u'proj3',
        'host_name': u'oss-fuzz-linux-zone3-host-efgh'
    }, {
        'instance_num': 3,
        'worker_name': u'oss-fuzz-linux-zone3-worker-proj3-0007',
        'project_name': u'proj3',
        'host_name': u'oss-fuzz-linux-zone3-host-efgh'
    }, {
        'instance_num': 4,
        'worker_name': u'oss-fuzz-linux-zone3-worker-proj3-0008',
        'project_name': u'proj3',
        'host_name': u'oss-fuzz-linux-zone3-host-efgh'
    }, {
        'instance_num': 5,
        'worker_name': u'oss-fuzz-linux-zone3-worker-proj3-0009',
        'project_name': u'proj3',
        'host_name': u'oss-fuzz-linux-zone3-host-efgh'
    }, {
        'instance_num': 6,
        'worker_name': u'oss-fuzz-linux-zone3-worker-proj4-0001',
        'project_name': u'proj4',
        'host_name': u'oss-fuzz-linux-zone3-host-efgh'
    }, {
        'instance_num': 7,
        'worker_name': u'oss-fuzz-linux-zone3-worker-proj5-0001',
        'project_name': u'proj5',
        'host_name': u'oss-fuzz-linux-zone3-host-efgh'
    }, {
        'instance_num': 0,
        'worker_name': u'oss-fuzz-linux-zone3-worker-high-end-proj6-0001',
        'project_name': u'proj6',
        'host_name': u'oss-fuzz-linux-zone3-host-high-end-1'
    }, {
        'instance_num': 1,
        'worker_name': u'oss-fuzz-linux-zone3-worker-high-end-proj6-0002',
        'project_name': u'proj6',
        'host_name': u'oss-fuzz-linux-zone3-host-high-end-1'
    }], [
        assignment.to_dict()
        for assignment in data_types.HostWorkerAssignment.query()
    ])


class OssFuzzDistributeCpusTest(unittest.TestCase):
  """Tests OSS-Fuzz CPU distribution."""

  def setUp(self):
    test_helpers.patch(self, [
        'clusterfuzz._internal.google_cloud_utils.compute_engine_projects.load_project',
    ])
    self.mock.load_project.return_value = OSS_FUZZ_CLUSTERS

  def test_equal(self):
    """Tests for each project receiving equal share."""
    projects = [
        data_types.OssFuzzProject(name='proj1', cpu_weight=1.0),
        data_types.OssFuzzProject(name='proj2', cpu_weight=1.0),
        data_types.OssFuzzProject(name='proj3', cpu_weight=1.0),
    ]

    result = manage_vms.OssFuzzClustersManager(
        'clusterfuzz-external').distribute_cpus(projects, 30)
    self.assertListEqual([10, 10, 10], result)

  def test_equal_uneven(self):
    """Tests for each project receiving equal share with an uneven division."""
    projects = [
        data_types.OssFuzzProject(name='proj1', cpu_weight=1.0),
        data_types.OssFuzzProject(name='proj2', cpu_weight=1.0),
        data_types.OssFuzzProject(name='proj3', cpu_weight=1.0),
    ]

    result = manage_vms.OssFuzzClustersManager(
        'clusterfuzz-external').distribute_cpus(projects, 31)
    self.assertListEqual([11, 10, 10], result)

    result = manage_vms.OssFuzzClustersManager(
        'clusterfuzz-external').distribute_cpus(projects, 32)
    self.assertListEqual([11, 11, 10], result)

  def test_weight_preference(self):
    """Tests that remainders are given to projects with higher weights

    first.
    """
    projects = [
        data_types.OssFuzzProject(name='proj1', cpu_weight=1.0),
        data_types.OssFuzzProject(name='proj2', cpu_weight=1.01),
        data_types.OssFuzzProject(name='proj3', cpu_weight=1.1),
    ]

    result = manage_vms.OssFuzzClustersManager(
        'clusterfuzz-external').distribute_cpus(projects, 4)
    self.assertListEqual([1, 1, 2], result)

    result = manage_vms.OssFuzzClustersManager(
        'clusterfuzz-external').distribute_cpus(projects, 5)
    self.assertListEqual([1, 2, 2], result)

  def test_not_enough(self):
    """Tests allocation with not enough CPUs."""
    projects = [
        data_types.OssFuzzProject(name='proj1', cpu_weight=1.0),
        data_types.OssFuzzProject(name='proj2', cpu_weight=1.0),
        data_types.OssFuzzProject(name='proj3', cpu_weight=1.0),
    ]

    result = manage_vms.OssFuzzClustersManager(
        'clusterfuzz-external').distribute_cpus(projects, 1)
    self.assertListEqual([1, 0, 0], result)

    result = manage_vms.OssFuzzClustersManager(
        'clusterfuzz-external').distribute_cpus(projects, 2)
    self.assertListEqual([1, 1, 0], result)

  def test_minimum(self):
    """Tests that projects are given a minimum share."""
    projects = [
        data_types.OssFuzzProject(name='proj1', cpu_weight=0.0),
        data_types.OssFuzzProject(name='proj2', cpu_weight=0.0),
        data_types.OssFuzzProject(name='proj3', cpu_weight=0.0),
    ]

    result = manage_vms.OssFuzzClustersManager(
        'clusterfuzz-external').distribute_cpus(projects, 3)
    self.assertListEqual([1, 1, 1], result)

    result = manage_vms.OssFuzzClustersManager(
        'clusterfuzz-external').distribute_cpus(projects, 10)
    self.assertListEqual([4, 3, 3], result)

  def test_maximum(self):
    """Tests that projects are capped at the maximum share."""
    projects = [
        data_types.OssFuzzProject(name='proj1', cpu_weight=1.0),
        data_types.OssFuzzProject(name='proj2', cpu_weight=1.0),
        data_types.OssFuzzProject(name='proj3', cpu_weight=1.0),
    ]

    result = manage_vms.OssFuzzClustersManager(
        'clusterfuzz-external').distribute_cpus(projects, 10000)
    self.assertListEqual([1000, 1000, 1000], result)

  def test_primes(self):
    """Test a bunch of different distributions."""
    projects = [
        data_types.OssFuzzProject(name='proj1', cpu_weight=2.0),
        data_types.OssFuzzProject(name='proj2', cpu_weight=3.0),
        data_types.OssFuzzProject(name='proj3', cpu_weight=5.0),
        data_types.OssFuzzProject(name='proj4', cpu_weight=7.0),
        data_types.OssFuzzProject(name='proj5', cpu_weight=11.0),
    ]

    result = manage_vms.OssFuzzClustersManager(
        'clusterfuzz-external').distribute_cpus(projects, 101)
    self.assertListEqual([7, 10, 18, 26, 40], result)
    self.assertEqual(101, sum(result))

    result = manage_vms.OssFuzzClustersManager(
        'clusterfuzz-external').distribute_cpus(projects, 887)
    self.assertListEqual([63, 95, 158, 222, 349], result)
    self.assertEqual(887, sum(result))

    result = manage_vms.OssFuzzClustersManager(
        'clusterfuzz-external').distribute_cpus(projects, 2741)
    self.assertListEqual([214, 313, 509, 705, 1000], result)
    self.assertEqual(2741, sum(result))

    result = manage_vms.OssFuzzClustersManager(
        'clusterfuzz-external').distribute_cpus(projects, 3571)
    self.assertListEqual([356, 483, 738, 994, 1000], result)
    self.assertEqual(3571, sum(result))


@test_utils.with_cloud_emulators('datastore')
class AssignHostWorkerTest(unittest.TestCase):
  """Tests host -> worker assignment."""

  def setUp(self):
    test_helpers.patch(self, [
        'clusterfuzz._internal.google_cloud_utils.compute_engine_projects.load_project',
    ])
    self.mock.load_project.return_value = OSS_FUZZ_CLUSTERS

  def test_assign_keep_existing(self):
    """Test that assignment keeps existing assignments."""
    host_names = ['host']
    worker_instances = [
        manage_vms.WorkerInstance(name='worker-proj-0', project='proj'),
        manage_vms.WorkerInstance(name='worker-proj-1', project='proj'),
        manage_vms.WorkerInstance(name='worker-proj-2', project='proj'),
        manage_vms.WorkerInstance(name='worker-proj-3', project='proj'),
        manage_vms.WorkerInstance(name='worker-proj-4', project='proj'),
        manage_vms.WorkerInstance(name='worker-proj-5', project='proj'),
        manage_vms.WorkerInstance(name='worker-proj-6', project='proj'),
        manage_vms.WorkerInstance(name='worker-proj-7', project='proj'),
    ]

    data_types.HostWorkerAssignment(
        host_name='host',
        instance_num=2,
        worker_name='worker-proj-6',
        project_name='proj',
        id='host-2').put()

    data_types.HostWorkerAssignment(
        host_name='host',
        instance_num=3,
        worker_name='worker-proj-1',
        project_name='proj',
        id='host-3').put()

    data_types.HostWorkerAssignment(
        host_name='host',
        instance_num=0,
        worker_name='worker-nonexistent-1',
        project_name='nonexistent',
        id='host-0').put()

    manager = manage_vms.OssFuzzClustersManager('clusterfuzz-external')
    new_assignments = manager.do_assign_hosts_to_workers(
        host_names, worker_instances, 8)
    self.assertListEqual([
        {
            'host_name': u'host',
            'instance_num': 0,
            'project_name': 'proj',
            'worker_name': 'worker-proj-0'
        },
        {
            'host_name': u'host',
            'instance_num': 1,
            'project_name': 'proj',
            'worker_name': 'worker-proj-2'
        },
        {
            'host_name': u'host',
            'instance_num': 4,
            'project_name': 'proj',
            'worker_name': 'worker-proj-3'
        },
        {
            'host_name': u'host',
            'instance_num': 5,
            'project_name': 'proj',
            'worker_name': 'worker-proj-4'
        },
        {
            'host_name': u'host',
            'instance_num': 6,
            'project_name': 'proj',
            'worker_name': 'worker-proj-5'
        },
        {
            'host_name': u'host',
            'instance_num': 7,
            'project_name': 'proj',
            'worker_name': 'worker-proj-7'
        },
    ], [assignment.to_dict() for assignment in new_assignments])



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for blobs."""

import os
import unittest

import mock

from clusterfuzz._internal.google_cloud_utils import blobs
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import test_utils

TEST_UUID = 'e612999f-ed89-4496-b4bd-3e8c7d8da18a'


@test_utils.with_cloud_emulators('datastore')
class BlobsTest(unittest.TestCase):
  """Tests for blobs."""

  def setUp(self):
    helpers.patch_environ(self)
    helpers.patch(self, [
        'clusterfuzz._internal.system.environment.is_running_on_app_engine',
        'clusterfuzz._internal.google_cloud_utils.blobs.generate_new_blob_name',
        'clusterfuzz._internal.google_cloud_utils.storage.copy_file_from',
        'clusterfuzz._internal.google_cloud_utils.storage.copy_file_to',
        'clusterfuzz._internal.google_cloud_utils.storage.delete',
        'clusterfuzz._internal.google_cloud_utils.storage.read_data',
        'clusterfuzz._internal.google_cloud_utils.storage.get',
    ])

    self.mock.is_running_on_app_engine.return_value = True
    os.environ['TEST_BLOBS_BUCKET'] = 'blobs-bucket'

    blobs.BlobInfo(
        id='legacyblobkey',
        filename='legacy-file',
        size=123,
        gs_object_name='/blobs-bucket/legacy').put()

    self.mock.get.return_value = {
        'metadata': {
            'filename': 'gcs-file',
        },
        'size': 456,
    }

    self.mock.is_running_on_app_engine.return_value = False
    self.mock.generate_new_blob_name.return_value = 'new-key'

    self.mock.copy_file_from.return_value = True
    self.mock.copy_file_to.return_value = True
    self.mock.read_data.return_value = b'data'
    self.mock.delete.return_value = True

  def test_get_gcs_path(self):
    """Test get_gcs_path."""
    self.assertEqual('/blobs-bucket/' + TEST_UUID,
                     blobs.get_gcs_path(TEST_UUID))

  def test_get_gcs_path_legacy(self):
    """Test get_gcs_path for legacy blobs."""
    self.assertEqual('/blobs-bucket/legacy',
                     blobs.get_gcs_path('legacyblobkey'))

  def test_get_blob_size_gcs(self):
    """Test get_blob_size for GCS blob."""
    self.assertEqual(456, blobs.get_blob_size(TEST_UUID))

  def test_get_blob_size_legacy(self):
    """Test get_blob_size for legacy blob."""
    self.assertEqual(123, blobs.get_blob_size('legacyblobkey'))

  def test_get_blob_size_invalid(self):
    """Test get_blob_size for invalid blob."""
    self.assertEqual(None, blobs.get_blob_size(None))
    self.assertEqual(None, blobs.get_blob_size('NA'))
    self.assertEqual(None, blobs.get_blob_size(''))

  def test_get_blob_info_gcs(self):
    """Test get_gcs_path for GCS files."""
    blob_info = blobs.get_blob_info(TEST_UUID)
    self.assertEqual(TEST_UUID, blob_info.key())
    self.assertEqual(456, blob_info.size)
    self.assertEqual('gcs-file', blob_info.filename)

  def test_get_blob_info_legacy(self):
    """Test get_gcs_path for legacy files."""
    blob_info = blobs.get_blob_info('legacyblobkey')
    self.assertEqual('legacyblobkey', blob_info.key())
    self.assertEqual(123, blob_info.size)
    self.assertEqual('legacy-file', blob_info.filename)

  def test_delete(self):
    """Test delete for GCS files."""
    self.assertEqual(True, blobs.delete_blob(TEST_UUID))
    self.mock.delete.assert_has_calls([
        mock.call('/blobs-bucket/' + TEST_UUID),
    ])

  def test_delete_legacy(self):
    """Test delete for legacy files."""
    self.assertEqual(True, blobs.delete_blob('legacyblobkey'))
    self.mock.delete.assert_has_calls([
        mock.call('/blobs-bucket/legacy'),
    ])

  def test_write_blob_file(self):
    """Test write_blob with a filename."""
    self.mock.get.return_value = None
    self.assertEqual('new-key', blobs.write_blob('/file'))
    self.mock.copy_file_to.assert_has_calls([
        mock.call(
            '/file', '/blobs-bucket/new-key', metadata={
                'filename': 'file',
            }),
    ])

  def test_write_blob_handle(self):
    """Test write_blob with a handle."""
    self.mock.get.return_value = None
    handle = mock.Mock()
    handle.name = 'filename'

    self.assertEqual('new-key', blobs.write_blob(handle))
    self.mock.copy_file_to.assert_has_calls([
        mock.call(
            handle, '/blobs-bucket/new-key', metadata={
                'filename': 'filename',
            }),
    ])

  def test_read_blob_to_disk(self):
    """Test read_blob_to_disk for GCS files."""
    self.assertTrue(blobs.read_blob_to_disk(TEST_UUID, '/file'))
    self.mock.copy_file_from.assert_has_calls([
        mock.call('/blobs-bucket/' + TEST_UUID, '/file'),
    ])

  def test_read_blob_to_disk_legacy(self):
    """Test read_blob_to_disk for legacy files."""
    self.assertTrue(blobs.read_blob_to_disk('legacyblobkey', '/file'))
    self.mock.copy_file_from.assert_has_calls([
        mock.call('/blobs-bucket/legacy', '/file'),
    ])

  def test_read_key(self):
    """Test read key for GCS files."""
    self.assertEqual(b'data', blobs.read_key(TEST_UUID))
    self.mock.read_data.assert_has_calls([
        mock.call('/blobs-bucket/' + TEST_UUID),
    ])

  def test_read_key_legacy(self):
    """Test read key for legacy files."""
    self.assertEqual(b'data', blobs.read_key('legacyblobkey'))
    self.mock.read_data.assert_has_calls([
        mock.call('/blobs-bucket/legacy'),
    ])
